/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * This is GNU GO, a Go program. Contact gnugo@gnu.org, or see   *
 * http://www.gnu.org/software/gnugo/ for more information.      *
 *                                                               *
 * Copyright 1999 and 2000 by the Free Software Foundation.      *
 *                                                               *
 * This program is free software; you can redistribute it and/or *
 * modify it under the terms of the GNU General Public License   *
 * as published by the Free Software Foundation - version 2.     *
 *                                                               *
 * This program is distributed in the hope that it will be       *
 * useful, but WITHOUT ANY WARRANTY; without even the implied    *
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR       *
 * PURPOSE.  See the GNU General Public License in file COPYING  *
 * for more details.                                             *
 *                                                               *
 * You should have received a copy of the GNU General Public     *
 * License along with this program; if not, write to the Free    *
 * Software Foundation, Inc., 59 Temple Place - Suite 330,       *
 * Boston, MA 02111, USA                                         *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


/* ---------------------------------------------------------------- *
 * gnugo.h
 *	This file contains the public interface to the GNU Go engine.
 * ---------------------------------------------------------------- */


#ifndef _GNUGO_H_
#define _GNUGO_H_


#include <stdio.h>


/* ================================================================ */
/*                some public macros used everywhere                */
/* ================================================================ */


/* Colors */
#define EMPTY        0
#define WHITE        1
#define BLACK        2
#define GRAY_BORDER  3
#define WHITE_BORDER 4
#define BLACK_BORDER 5
#define NONE         6		/* for use with is_computer_player */

#define OTHER_COLOR(color)  (WHITE+BLACK-(color))


/* Group statuses */
#define DEAD      0
#define ALIVE     1
#define CRITICAL  2 
#define UNKNOWN   3
#define UNCHECKED 4
#define MAX_DRAGON_STATUS 4	/* used to size an array in matchpat.c */


/* 
 * It is assumed in reading a ladder if stackp >= depth that
 * as soon as a bounding stone is in atari, the string is safe.
 * It is used similarly at other places in reading.c to implement simplifying
 * assumptions when stackp is large. DEPTH is the default value of depth.
 *
 * Unfortunately any such scheme invites the ``horizon effect.'' Increasing
 * DEPTH will make the program stronger and slower.
 *
 */

#define DEPTH           14
#define BACKFILL_DEPTH   9
#define FOURLIB_DEPTH    5
#define KO_DEPTH         8

#define LIFE_EYESIZE     6

/* style definitions */
#define STY_DEFAULT          1       /* standard fuseki, default */
#define STY_NO_FUSEKI        2       /* no fuseki, just plays 4 corners */
#define STY_TENUKI           8       /* more tenuki, compatible with default */
#define STY_FEARLESS        16       /* fearless, compatible with default */


/* ================================================================ */
/*                        Board manipulation                        */
/* ================================================================ */


/* Board sizes */
#define MIN_BOARD     3		/* minimum supported board size */
#define MAX_BOARD    21         /* maximum supported board size */


#define MAX_HANDICAP  9		/* maximum supported handicap   */


/* This type is used to store each piece on the board.
 *
 * On a 486, char is best, since the time taken to push and pop
 * becomes significant otherwise. On other platforms, an int may
 * be better, e.g. if memcpy() is particularly fast, or if
 * character access is very slow.
 */

typedef unsigned char Intersection;


typedef struct {
  int          boardsize;
  Intersection board[MAX_BOARD][MAX_BOARD];
  int          ko_i;
  int          ko_j;

  int          white_captured;
  int          black_captured;
} Position;

void position_clear(Position *pos, int boardsize);
void position_copy(Position *to, Position *from);
void position_add_stone(Position *pos, int i, int j, int color);
void position_remove_stone(Position *pos, int i, int j);
void position_play_move(Position *pos, int i, int j, int color);
int  position_legal(Position *pos, int i, int j, int color);

int  position_placehand(Position *pos, int handicap);
void position_recordhand(Position *pos, int change_sgf_tree);
int  position_sethand(Position *pos, int handicap, int change_sgf_tree);

int  position_genmove(Position *pos, int *i, int *j, int color,
		      int move_number);

int  position_attack(Position *pos, int m, int n, int *i, int *j);
int  position_find_defense(Position *pos, int m, int n, int *i, int *j);

void position_who_wins(Position *pos, int color, float komi, FILE *outfile);
void position_evaluate_territory(Position *pos, int *white, int *black);


/* ================================================================ */
/*                           Game handling                          */
/* ================================================================ */


typedef struct gameinfo {
  int       handicap;
  float     komi;

  int       move_number;
  Position  position;
  int       to_move;		/* whose move it currently is */

  int       seed;		/* random seed */
  int       computer_player;	/* BLACK, WHITE, or EMPTY (used as BOTH) */
  char      outfile[128];
} Gameinfo;

void ginfo_clear(Gameinfo *ginfo, int boardsize);
void ginfo_print(Gameinfo *ginfo);


/* ================================================================ */
/*                           global variables                       */
/* ================================================================ */


/* Miscellaneous debug options. */
extern int verbose;		/* Bore the opponent. */
extern int allpats;		/* generate all patterns, even small ones */
extern int printworms;		/* print full data on each string */
extern int printmoyo;		/* print moyo board each move */
extern int printdragons;	/* print full data on each dragon */
extern int printboard;		/* print board each move */
extern int showstack;		/* debug stack pointer */
extern int showstatistics;	/* print statistics */
extern int profile_patterns;	/* print statistics of pattern usage */


/* debug flag bits */
/* NOTE : can specify -d0x... */
#define DEBUG_GENERAL      0x0001 /* unused */
#define DEBUG_COUNT        0x0002 /* obsolete */
#define DEBUG_BOARD        0x0004 /* obsolete */
#define DEBUG_CAPTURE      0x0008 /* obsolete */
#define DEBUG_STACK        0x0010 /* obsolete */
#define DEBUG_INFLUENCE    0x0020
#define DEBUG_HELPER       0x0040
#define DEBUG_LOADSGF      0x0080
#define DEBUG_WORMS        0x0100 /* unused */
#define DEBUG_LADDER       0x0200
#define DEBUG_MATCHER      0x0400
#define DEBUG_DEFENDER     0x0800 /* unused */
#define DEBUG_ATTACKER     0x1000 /* unused */
#define DEBUG_BORDER       0x2000
#define DEBUG_DRAGONS      0x4000
#define DEBUG_SAVESGF      0x8000 /* unused */
#define DEBUG_ESCAPE       0x10000
#define DEBUG_SEMEAI       0x20000
#define DEBUG_EYES         0x40000
#define DEBUG_MOVE_REASONS 0x80000 /* unused */
#define DEBUG_MOUSE        0x100000

/* hash flag bits */
#define HASH_FIND_DEFENSE 0x0001  /* NOTE : can specify -d0x... */
#define HASH_DEFEND1      0x0002
#define HASH_DEFEND2      0x0004
#define HASH_DEFEND3      0x0008
#define HASH_ATTACK       0x0010
#define HASH_ATTACK2      0x0020
#define HASH_ATTACK3      0x0040
#define HASH_DEFEND4      0x0080
#define HASH_ALL          0xffff
#define HASH_DEFAULT      HASH_ATTACK
#define HASH_NOTHING      0


extern int debug;		/* debug flags */
extern int hashflags;		/* hash flags */
extern int style;		/* playing style */
extern int mouse;               /* use mouse code */
extern int life;                /* use eye reading code */
extern int life_eyesize;
extern int fusekidb;            /* use fuseki database */
extern int hurry;		/* play faster */
extern int terse;		/* terse output */
extern int showtime;		/* print genmove time */
extern int analyzerflag;

/* Reading parameters */
extern int nominal_depth;	   /* deep reading cutoff, nominal value */
extern int nominal_backfill_depth; /* deep reading cutoff, nominal value */
extern int nominal_fourlib_depth;  /* deep reading cutoff, nominal value */
extern int nominal_ko_depth;	   /* deep ko reading cutoff, nominal value */

/* The current position */
extern int movenum;          /* movenumber */

extern int potential_moves[MAX_BOARD][MAX_BOARD];

extern int terri_eval[3];
extern int moyo_eval[3];

extern char *analyzerfile;

extern volatile int time_to_die;   /* set by signal handlers */
extern int color_has_played;  /* whether color has placed a stone yet */

/* ================================================================ */
/*                 tracing and debugging functions                  */
/* ================================================================ */


/* Colors. */
#define COLOR_BLACK   0
#define COLOR_RED     1
#define COLOR_GREEN   2
#define COLOR_YELLOW  3
#define COLOR_BLUE    4
#define COLOR_MAGENTA 5
#define COLOR_CYAN    6
#define COLOR_WHITE   7

/* showbord.c */
void start_draw_board(void);
void draw_color_char(int m, int n, int c, int color);
void draw_char(int m, int n, int c);
void end_draw_board(void);
void showboard(int xo);  /* ascii rep. of board to stdout */

/* utils.c */
void gprintf(const char *fmt, ...);
void mprintf(const char *fmt, ...);
const char *color_to_string(int color);
double gg_gettimeofday(void);


/* influence.c */
void debug_influence_move(int i, int j);

#ifdef __GNUC__

/* gnuc allows variadic macros, so the tests can be done inline */
#define TRACE(fmt, args...) \
    do { if (verbose) gprintf(fmt, ##args); } while(0)
#define RTRACE(fmt, args...) \
    do { if (verbose >= 3) gprintf(fmt, ##args); } while(0)
#define VTRACE(fmt, args...) \
    do { if (verbose >= 4) gprintf(fmt, ##args); } while(0)
#define DEBUG(level, fmt, args...) \
    do { if ((debug & (level))) gprintf(fmt, ##args); } while(0)

#else

/* else we need fns to do these */
void TRACE(const char *fmt, ...);
void RTRACE(const char *fmt, ...);
void VTRACE(const char *fmt, ...);
void DEBUG(int level, const char *fmt, ...);

#endif

/* genmove.c */
#define EXAMINE_WORMS               1
#define EXAMINE_INITIAL_INFLUENCE   2
#define EXAMINE_DRAGONS             3
#define EXAMINE_MOUSE_REASONS       4
#define EXAMINE_MOYO                5
#define EXAMINE_ESCAPE_POTENTIAL    6
#define EXAMINE_INITIAL_INFLUENCE2  7

#define EXAMINE_ALL                 99

void examine_position(int color, int how_much);
void initialize_engine(void);


/* ================================================================ */
/*                           public functions                       */
/* ================================================================ */


/* These functions are the official ones which are supported
 * by the engine for use by the user interfaces.
 */

/* board.c */
void init_gnugo(float memory);
void init_board(void);
void popgo(void);
int trymove(int i, int j, int color, const char *message, int k, int l);
int tryko(int i, int j, int color, const char *message);
void add_stone(int i, int j, int color); /* FIXME: Move to liberty.h? */
void remove_stone(int i, int j);
void remove_string(int i, int j);
void remove_string_and_mark(int i, int j, char mw[MAX_BOARD][MAX_BOARD], char mark);
void play_move(int i, int j, int color);
int legal(int i, int j, int color);
/* int approxlib(int m, int n, int color, int maxlib) */
/* int is_ko(int i, int j, int color) */

/* utils.c */
void change_dragon_status(int x, int y, int status);
void count_territory( int *white, int *black);
void evaluate_territory( int *white, int *black);
int same_worm(int ai, int aj, int bi, int bj);
int same_dragon(int ai, int aj, int bi, int bj);
int is_worm_origin(int wi, int wj, int i, int j);
void decideeye(int m, int n);


/* high-level routine to generate the best move for given color */
int genmove(int *i, int *j, int color);

/* Basic information gathering. */
int make_worms(void);
void make_dragons(int color);
void show_dragons(void);
int dragon_status(int i, int j);


/* data concerning moyo */
extern int terri_test[3];
extern int moyo_test[3];
extern int very_big_move[3];

/* moyo functions */
int make_moyo(int color);
void print_moyo(int color);

/* FIXME: This is in moyo.c, but should be somewhere else. */
void who_wins(int color, float fkomi, FILE* stdwhat);

/* debugging functions */
void prepare_pattern_profiling(void);
void report_pattern_profiling(void);

/* sgffile.c */
void sgffile_move_made(int i, int j, int color, int value);
void sgffile_put_stone(int i, int j, int color);

int sgffile_open_file(const char *);
int sgffile_close_file(void);
int sgffile_write_gameinfo(Gameinfo *gameinfo, const char *);
void sgffile_write_comment(const char *comment);
void sgffile_printboard(int next);

void sgffile_decidestring(int i, int j, const char *sgf_output);
void sgffile_decidedragon(int i, int j, const char *sgf_output);
void sgffile_decideposition(int color, const char *sgf_output);

#endif  /* _GNUGO_H_ */


/*
 * Local Variables:
 * tab-width: 8
 * c-basic-offset: 2
 * End:
 */
