/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * This is GNU GO, a Go program. Contact gnugo@gnu.org, or see   *
 * http://www.gnu.org/software/gnugo/ for more information.      *
 *                                                               *
 * Copyright 1999 and 2000 by the Free Software Foundation.      *
 *                                                               *
 * This program is free software; you can redistribute it and/or *
 * modify it under the terms of the GNU General Public License   *
 * as published by the Free Software Foundation - version 2.     *
 *                                                               *
 * This program is distributed in the hope that it will be       *
 * useful, but WITHOUT ANY WARRANTY; without even the implied    *
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR       *
 * PURPOSE.  See the GNU General Public License in file COPYING  *
 * for more details.                                             *
 *                                                               *
 * You should have received a copy of the GNU General Public     *
 * License along with this program; if not, write to the Free    *
 * Software Foundation, Inc., 59 Temple Place - Suite 330,       *
 * Boston, MA 02111, USA                                         *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


/*
 * This file is about life & death.
 * 
 * For now, we only consider eye spaces.  Code for extending and
 * shrinking eye spaces is in mouse.c. 
 *
 * We will only handle small eye spaces for now (<= 6 during testing).
 * At these small sizes, we make the simplifying assumption that
 * if any part of the surrounding wall is captured, the resulting
 * eye space will lead to 0 eyes.  This assumtion can be dealt with
 * later when the basic code works.
 *
 * Strategy:
 * Try to fill the entire eye space until there are only eyes of
 * size 1 left. Then nobody can play there, and we can count the eyes.
 */

/* FIXME: TODO
 *  - Handle ko
 *  - Handle seki
 *  - Try to find out how many outside liberties are needed for the optimal
 *    sequence.
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "liberty.h"
#include "location_set.h"


#define MAX_EYE_SIZE  10

static int minimize_eyes(int color, int *min, int *attacki, int *attackj, 
			 struct eye_data eyedata[MAX_BOARD][MAX_BOARD],
			 Location_set *eye,
			 Location_set *diag_defends, Location_set *diag_attacks,
			 int last_move_was_pass, Location_set *moves);
static int maximize_eyes(int color, int *max, int *attacki, int *attackj, 
			 struct eye_data eyedata[MAX_BOARD][MAX_BOARD],
			 Location_set *eye,
			 Location_set *diag_defends, Location_set *diag_attacks,
			 int last_move_was_pass, Location_set *moves);
static void get_boundary(Location_set *boundary, 
			 struct eye_data eyedata[MAX_BOARD][MAX_BOARD],
			 Location_set *eyepoints, int color);
static void life_showboard(void);

/* Used by life_showboard(). */
static int stackp_when_called = 0;


/* ================================================================ */
/*                   Hash table for recognize_eye.                  */
/* ================================================================ */


/*
 * Layout of the hash table:
 *
 *   Each entry is one 32 bit integer. (FIXME: Don't waste 32 bits on 
 *   64 bit machines.) 
 *
 *   byte 0 (LSB):  Result byte (see below)
 *   byte 1:        Where the attacker should play (index in the set)
 *   byte 2:        Where the attacker should play.
 *   byte 3 (MSB):  Reserved
 *
 *   The result byte:
 *   +-------------------------------+
 *   | 7 | 6 | 5 | 4 | 3 | 2 | 1 | 0 |
 *   +-------------------------------+
 *   |vis|xxx|  dresult  |  aresult  |
 *   +-------------------------------+
 *
 *   aresult and dresult:
 *   0:   No result known yet.  This is used during the search.
 *   1:   0 eyes
 *   2:   1 eye
 *   3:   2 eyes
 *   4-7: Reserved
 */

/* defines for the entire entry */
#define ATTACK_MOVE_SHIFT  8
#define ATTACK_BYTE        (0xff << ATTACK_MOVE_SHIFT) /* bit mask */
#define ATTACK_NO_MOVE     (255 << ATTACK_MOVE_SHIFT)  /* value 255 */
#define DEFENSE_MOVE_SHIFT 16
#define DEFENSE_BYTE       (0xff << DEFENSE_MOVE_SHIFT) /* bit mask */
#define DEFENSE_NO_MOVE    (255 << DEFENSE_MOVE_SHIFT)

/* defines for the result byte */
#define VISITED              0x80	/* Set if this node is visited */
#define RESULT_MASK          0x07
#define ATTACK_RESULT_SHIFT  0
#define DEFENSE_RESULT_SHIFT 3


/* The hash table will not have to handle any collisions, since it
 * is complete. 
 */
#define EYE_HASH_SIZE  (3*3*3*3*3*3*3*3*3*3*2) /* 3^MAX_EYE_SIZE*2 */


static int  eyehash_table[EYE_HASH_SIZE];


static void  eyehash_clear(void);
static int   eyehash_index(Location_set *eyepoints, int to_move);


/*
 * Clear the hash table.
 */

static void
eyehash_clear()
{
  memset(eyehash_table, 0, sizeof(eyehash_table));
}


/*
 * Return the index to the hash entry for the current position, and the
 * side to move.
 *
 * FIXME: This has to be changed when we take into account ko.
 */

static int
eyehash_index(Location_set *eyepoints, int to_move)
{
  Location_set_iterator   iter;
  int  hashval;
  int  i, j;

  /* Step through all locations in `eyepoints'. */
  hashval = 0;
  for (ls_iter_init(eyepoints, &iter);
       ls_iter_next(&iter, &i, &j);
       ls_iter_update(&iter))
    hashval = hashval * 3 + p[i][j];
  hashval = hashval * 2 + (to_move - 1);

  return hashval;
}


/*
 * Get the attacking move from the cell at `hashindex'.  
 *
 * The move is coded as the nth move within `eyepoints'.  The NULL move
 * (a pass) is coded as 255.
 */

static void
eyehash_get_attack_move(Location_set *eyepoints, int hashindex,
			int *attacki, int *attackj)
{
  Location_set_iterator  iter;
  int  move_index;
  int  m, n;
  int  i;

  move_index = (eyehash_table[hashindex] >> ATTACK_MOVE_SHIFT) & 255;

  if (move_index == 255) {
    m = -1;
    n = -1;
  } else
    for (ls_iter_init(eyepoints, &iter), i = 0;
	 ls_iter_next(&iter, &m, &n), i < move_index;
	 ls_iter_update(&iter), i++)
      /* NOTHING */;

  if (attacki) *attacki = m;
  if (attackj) *attackj = n;
}


/*
 * Set the attacking move in the cell at `hashindex' to (m, n).
 */

static void
eyehash_set_attack_move(Location_set *eyepoints, int hashindex,
			int attacki, int attackj)
{
  Location_set_iterator  iter;
  int  m,n;
  int  i;

  if (attacki == -1)
    eyehash_table[hashindex] |= ATTACK_NO_MOVE;
  else
    for (ls_iter_init(eyepoints, &iter), i = 0;
	 ls_iter_next(&iter, &m, &n);
	 ls_iter_update(&iter), i++)
      if (m == attacki && n == attackj) {
	/* We don't have to mask away the previous entry because we
	 * never change anything.
	 */
	eyehash_table[hashindex] |= i << ATTACK_MOVE_SHIFT;
	break;
      }
}


/*
 * Get the defending move from the cell at `hashindex'.  
 */

static void
eyehash_get_defense_move(Location_set *eyepoints, int hashindex,
			 int *defendi, int *defendj)
{
  Location_set_iterator  iter;
  int  move_index;
  int  m, n;
  int  i;

  move_index = (eyehash_table[hashindex] >> DEFENSE_MOVE_SHIFT) & 255;

  if (move_index == 255) {
    m = -1;
    n = -1;
  } else
    for (ls_iter_init(eyepoints, &iter), i = 0;
	 ls_iter_next(&iter, &m, &n), i < move_index;
	 ls_iter_update(&iter), i++)
      /* NOTHING */;

  if (defendi) *defendi = m;
  if (defendj) *defendj = n;
}


/*
 * Set the defending move in the cell at `hashindex' to (m, n).
 */

static void
eyehash_set_defense_move(Location_set *eyepoints, int hashindex,
			 int defendi, int defendj)
{
  Location_set_iterator  iter;
  int  m,n;
  int  i;

  if (defendi == -1)
    eyehash_table[hashindex] |= DEFENSE_NO_MOVE;
  else
    for (ls_iter_init(eyepoints, &iter), i = 0;
	 ls_iter_next(&iter, &m, &n);
	 ls_iter_update(&iter), i++)
      if (m == defendi && n == defendj) {
	eyehash_table[hashindex] |= i << DEFENSE_MOVE_SHIFT;
	break;
      }
}


/* ================================================================ */
/*                          Eye recognition                         */
/* ================================================================ */


/* The boundary of the eye we are investigating. */
static Location_set  boundary;


/*
 * Compute the maximum and minimum number of eyes reachable from the
 * eye space at (m, n). The maximum is defined for color to play
 * first, and the minimum is defined for other_color to play first.
 *
 * If max != min, (*attacki, *attackj) is set to the move which
 * reduces the eye and (*defensei, *defensej) is set to the move which
 * defends it. Most of the time the attack point and the defense point
 * will be the same.
 */

void
recognize_eye2(int m, int  n, 
	       int *attacki, int *attackj, int *defendi, int *defendj,
	       int *max, int *min, 
	       struct eye_data eyedata[MAX_BOARD][MAX_BOARD],
 	       int add_moves, int color)
{
  int eye_color, other;
  int i, j;
  int k;
  Location_set  eyepoints;
  Location_set           vital_moves;
  Location_set_iterator  iter;
  Location_set  diag_defends, diag_attacks;
  int   result1, result2;
  int   max1, min1;
  int   attacki1, attackj1;
  int   defendi1, defendj1;
  int  save_stackp;

  save_stackp = stackp;

  /* For use by debug outputs. */
  stackp_when_called = stackp;
  
  /* Set `eye_color' to the owner of the eye. */
  eye_color = eyedata[m][n].color;
  if (eye_color == BLACK_BORDER)
    eye_color = BLACK;
  if (eye_color == WHITE_BORDER)
    eye_color = WHITE;
  other = OTHER_COLOR(eye_color);

  /* Collect all the points of the eye into the set. */
  location_set_clear(&eyepoints);
  location_set_clear(&diag_defends);
  location_set_clear(&diag_attacks);
  for (i = 0; i < board_size; ++i)
    for (j = 0; j < board_size; ++j) {
      if (eyedata[i][j].origini == m && eyedata[i][j].originj == n) {
	location_set_add(&eyepoints, i, j);
	if (halfeye(i, j)) {
	  for (k=0; k<half_eye[i][j].num_defends; k++)
	    location_set_add(&diag_defends,
			     half_eye[i][j].ai[k], half_eye[i][j].aj[k]);
	  for (k=0; k<half_eye[i][j].num_attacks; k++)
	    location_set_add(&diag_attacks,
			     half_eye[i][j].di[k], half_eye[i][j].dj[k]);
	}
      }    
    }

  /* Get the boundary. */
  get_boundary(&boundary, eyedata, &eyepoints, eye_color);

  /* Clear the hash table for this eye. */
  eyehash_clear();

  if (0) {
    fprintf(stderr, "Eye (%s): ", color_to_string(eye_color));
    location_set_print(stderr, &eyepoints);
  }

  result1 = minimize_eyes(eye_color, &min1, &attacki1, &attackj1, 
			  eyedata, &eyepoints, &diag_defends, &diag_attacks, 0,
			  (color != eye_color && add_moves)
			    ? &vital_moves : NULL);
  result2 = maximize_eyes(eye_color, &max1, &defendi1, &defendj1, 
			  eyedata, &eyepoints, &diag_defends, &diag_attacks, 0,
			  (color == eye_color && add_moves)
			    ? &vital_moves : NULL);

  *min = min1;
  *max = max1;
  if (min1 != max1) {
    if (attacki) *attacki = attacki1;
    if (attackj) *attackj = attackj1;
    if (defendi) *defendi = defendi1;
    if (defendj) *defendj = defendj1;
    DEBUG(DEBUG_EYES, "  vital points: %m (attack)\n", *attacki, *attackj);
    DEBUG(DEBUG_EYES, "  vital points: %m (defense)\n", *defendi, *defendj);
    if (add_moves) {
      for (ls_iter_init(&vital_moves, &iter);
	   ls_iter_next(&iter, &i, &j);
	   ls_iter_update(&iter))
	{
	  add_vital_eye_move(i, j, m, n, eye_color);
	}
    }
  }

  assert(stackp == save_stackp);
}


/*
 * We have a set of locations in eyepoints forming an eye shape.
 * Check min number of eyes we can get if the opponent moves first.
 * `eye_color' is the owner of the eye.
 *
 * Return values:
 *   0: Ok. Min eyes in *min.
 *  -1: No result.  A cycle in the reading.
 */

static int
minimize_eyes(int eye_color, int *min, int *attacki, int *attackj, 
	      struct eye_data eyedata[MAX_BOARD][MAX_BOARD],
	      Location_set *eyepoints, 
	      Location_set *diag_defends, Location_set *diag_attacks,
	      int last_move_was_pass, Location_set *moves)
{
  int  other = OTHER_COLOR(eye_color);
  int  i, j;
  int  m, n;
  Location_set attacks;
  Location_set_iterator   iter;
  Location_set_iterator   iter2;
  int  num_other;
  int  num_legal_moves;
  int  num_eyes;
  int  bordercolor;
  int  eyesize;
  int  max1;
  int  localmax;
  int  hashindex;
  int  result;
  int  attacki1, attackj1;
  int  localattacki = -1;
  int  localattackj = -1;

  int  save_stackp;

  TRACE("entering minimize_eyes: stackp = %d\n", stackp);

  /* Collect attacking moves, union of eyepoints and diagonal vital moves */
  location_set_copy(&attacks, eyepoints);
  location_set_union(&attacks, diag_attacks);

  /* Check the hash table and see if we have been here before. */
  /* FIXME: The hashindex could be calculated incrementally.  Later. */
  hashindex = eyehash_index(&attacks, other);
  if (eyehash_table[hashindex] != 0) {
    result = (eyehash_table[hashindex] >> ATTACK_RESULT_SHIFT) & RESULT_MASK;
    if (result == 0) {
      /* No result known yet.  We have come back to the same position
       * again within the reading.  Just return and hope that we get a
       * better result on some other branch. 
       */
      TRACE("exiting minimize_eyes directly because of a reading loop.\n");
      return -1;
    }
    else if (result <= 3) {
      *min = result - 1;
      eyehash_get_attack_move(&attacks, hashindex, attacki, attackj);

      TRACE("exiting minimize_eyes - got result %d eyes from the table.\n",
	    *min);
      return 0;
    }
  }

  if (verbose)
    life_showboard();

  /* Collect how many moves the attacking color can do and how many stones
   * he has within the eye.
   */
  num_other = 0;
  num_legal_moves = 0;
  num_eyes = 0;
  for (ls_iter_init(&attacks, &iter);
       ls_iter_next(&iter, &i, &j);
       ls_iter_update(&iter))
  {
    if (p[i][j] == eye_color)
      continue;

    if (p[i][j] == other) {
      num_other++;
      continue;
    } 

    if (legal(i, j, other))
      num_legal_moves++;
    else {
      bordercolor = find_border(i, j, NULL, &eyesize);
      if (eyesize == 1 && !eyedata[i][j].marginal)
	num_eyes++;
    } 
  }

  /* Mark this node as visited, but without result so far. */
  eyehash_table[hashindex] |= VISITED;

  /* Check if we are done. This means that there are no stones of
   * the attacking color and that all the eye points are illegal
   * to move into. The only way this could happen is that they are
   * of size 1 and without possibility to capture anything.
   */
  if (num_other == 0 && num_legal_moves == 0) {
    if (num_eyes > 2)
      num_eyes = 2;
    *min = num_eyes;
    eyehash_table[hashindex] |= ATTACK_NO_MOVE;
    eyehash_table[hashindex] |= (num_eyes + 1) << ATTACK_RESULT_SHIFT;

    TRACE("exiting minimize_eyes\n");
    return 0;
  } 

  /* Start by see if doing nothing is enough. */
  localmax = 9999;
  if (!last_move_was_pass) {
    result = maximize_eyes(eye_color, &max1, &attacki1, &attackj1, 
			  eyedata, eyepoints, diag_defends, diag_attacks,
			  1, NULL);
    if (result == 0) {
      if (max1 < localmax) {
	localmax = max1;
	localattacki = -1;
	localattackj = -1;
      }

      /* Check if the pass yielded 0 eyes.  If so, we can as well
       * return here. */
      if (localmax == 0) {
	*min = 0;
	if (attacki) *attacki = -1;
	if (attackj) *attackj = -1;
	eyehash_table[hashindex] |= ATTACK_NO_MOVE;
	eyehash_table[hashindex] |= (0 + 1) << ATTACK_RESULT_SHIFT;

	TRACE("exiting minimize_eyes - a pass sufficed\n");
	return 0;
      }
    }  
  }

  /* Now try all possible moves and see what we get. */
  for (ls_iter_init(&attacks, &iter);
       ls_iter_next(&iter, &i, &j);
       ls_iter_update(&iter))
  {
    int  boundary_captured;

    if (p[i][j] == eye_color)
      continue;

    if (p[i][j] == other)
      continue; 

    /* Try the move and see if we can reduce the eyes. */
    save_stackp = stackp;
    boundary_captured = 0;
    TRACE("minimize_eyes: trymove at %m\n", i, j);
    if (trymove(i, j, other, "minimize_eyes", -1, -1)) {

      /* Check if any part of the boundary got captured.  If so, we assume
       * that we can get no eyes.
       *
       * FIXME: This is ok for now, but is it ok forever?
       */
      for (ls_iter_init(&boundary, &iter2);
	   ls_iter_next(&iter2, &m, &n);
	   ls_iter_update(&iter2))
      {
	if (p[m][n] == EMPTY) {
	  boundary_captured = 1;
	  break;
	}
      }

      if (!boundary_captured) {
	/* No boundary string was captured. The attacker  has made his move.
	 * Now let's answer him and see how many eyes we can get.
	 */
	result = maximize_eyes(eye_color, &max1, &attacki1, &attackj1, 
			  eyedata, eyepoints, diag_defends, diag_attacks,
			  0, NULL);
	if (result == 0) {
	  if (max1 < localmax) {
	    localmax = max1;
	    localattacki = i;
	    localattackj = j;
 	    if (moves)
 	      location_set_clear(moves);
 	  } 
 	  if (max1 == localmax && moves)
 	    location_set_add(moves, i, j);
	}
      }

      popgo();
    }
    assert(stackp == save_stackp);

    /* Check if the boundary was captured or the move yielded 0 eyes.
     * If so, we can as well return here. 
     */
#if 0
    if (boundary_captured || localmax == 0) {
      *min = 0;
      if (attacki) *attacki = i;
      if (attackj) *attackj = j;
      
      eyehash_set_attack_move(attacks, hashindex, i, j);
      eyehash_table[hashindex] |= (0 + 1) << ATTACK_RESULT_SHIFT;
      
      TRACE("exiting minimize_eyes\n");
      return 0;
    }
#endif
    
    /* We really want all eye minimizing moves */
    if (boundary_captured) {
      if (localmax) {
	localmax = 0;
	localattacki = i;
	localattackj = j;
	if (moves)
	  location_set_clear(moves);
      }
      if (moves)
	location_set_add(moves, i, j);
    }

  }

  /* If we didn't find anywhere to play, see how many eye spaces there were. */
  if (localmax == 9999) {
    localmax = num_eyes;
    if (localmax > 2)
      localmax = 2;
    *min = localmax;
    if (attacki) *attacki = -1;
    if (attackj) *attackj = -1;

    eyehash_table[hashindex] |= ATTACK_NO_MOVE;
    eyehash_table[hashindex] |= (localmax + 1) << ATTACK_RESULT_SHIFT;

    TRACE("exiting minimize_eyes\n");
    return 0;
  }

  if (localmax > 2)
    localmax = 2;

  *min = localmax;
  if (attacki) *attacki = localattacki;
  if (attackj) *attackj = localattackj;

  eyehash_set_attack_move(&attacks, hashindex, localattacki, localattackj);
  eyehash_table[hashindex] |= (localmax + 1) << ATTACK_RESULT_SHIFT;

  TRACE("exiting minimize_eyes\n");
  return 0;
}


/*
 * We have a set of locations in eyepoints forming an eye shape.
 * Check max number of eyes we can get if the opponent moves first.
 * `eye_color' is the owner of the eye.
 *
 * Return values:
 *   0: Ok. Max eyes in *max.
 *  -1: No result.  A cycle in the reading.
 */

static int
maximize_eyes(int eye_color, int *max, int *defendi, int *defendj, 
	      struct eye_data eyedata[MAX_BOARD][MAX_BOARD],
	      Location_set *eyepoints,
	      Location_set *diag_defends, Location_set *diag_attacks,
	      int last_move_was_pass, Location_set *moves)
{
  int  other = OTHER_COLOR(eye_color);
  int  i, j;
  Location_set defends;
  Location_set_iterator   iter;
  int  bordercolor;
  int  eyesize;
  int  min1;
  int  localmin;
  int  result;
  int  hashindex;
  int  defendi1, defendj1;
  int  localdefendi = -1;
  int  localdefendj = -1;

  int  save_stackp;

  TRACE("entering maximize_eyes: stackp = %d\n", stackp);

  /* Collect defending moves, union of eyepoints and diagonal vital moves */
  location_set_copy(&defends, eyepoints);
  location_set_union(&defends, diag_defends);

  /* Check the hash table and see if we have been here before. */
  /* FIXME: The hashindex could be calculated incrementally.  Later. */
  hashindex = eyehash_index(&defends, eye_color);
  if (eyehash_table[hashindex] != 0) {
    result = (eyehash_table[hashindex] >> DEFENSE_RESULT_SHIFT) & RESULT_MASK;
    if (result == 0) {
      /* No result known yet.  We have come back to the same position
       * again within the reading.  Just return and hope that we get a
       * better result on some other branch. 
       */
      TRACE("exiting maximize_eyes directly because of a reading loop.\n");
      return -1;
    }
    else if (result <= 3) {
      *max = result - 1;
      if ((eyehash_table[hashindex] & DEFENSE_BYTE) != DEFENSE_NO_MOVE)
	eyehash_get_defense_move(&defends, hashindex, defendi, defendj);

      TRACE("exiting maximize_eyes - got result %d eyes from the table.\n",
	    *max);
      return 0;
    }
  }

  if (verbose)
    life_showboard();

  localmin = 0;

  /* Mark this node as visited, but without result so far. */
  eyehash_table[hashindex] |= VISITED;

  /* Start by see if doing nothing is enough. */
  if (!last_move_was_pass) {
    result = minimize_eyes(eye_color, &min1, &defendi1, &defendj1, 
			  eyedata, eyepoints, diag_defends, diag_attacks,
			  1, NULL);
    if (result == 0) {
      if (min1 > localmin) {
	localmin = min1;
	localdefendi = -1;
	localdefendj = -1;
      }
    }
  }

  /* Now try all possible moves and see what we get. */
  for (ls_iter_init(&defends, &iter);
       ls_iter_next(&iter, &i, &j);
       ls_iter_update(&iter))
  {
    if (p[i][j] == eye_color)
      continue;

    if (p[i][j] == other)
      continue;

    /* If the eye is of size 1, and none of the neighbouring strings
     * is in atari, then this is a proper eye. Don't play there.
     */
    bordercolor = find_border(i, j, NULL, &eyesize);
    if (eyesize == 1) {
      if ((i == 0 
	   || (p[i-1][j] == eye_color
	       && countlib(i-1, j) > 1))
	  && (i == board_size-1
	      || (p[i+1][j] == eye_color
		  && countlib(i+1, j) > 1))
	  && (j == 0
	      || (p[i][j-1] == eye_color
		  && countlib(i, j-1) > 1))
	  && (j == board_size-1
	      || (p[i][j+1] == eye_color
		  && countlib(i, j+1) > 1)))
	continue;
    }

    /* Try the move and see if we can keep the eyes. */
    save_stackp = stackp;
    TRACE("maximize_eyes: trymove at %m\n", i, j);
    if (trymove(i, j, eye_color, "maximize_eyes", -1, -1)) {

      /* Ok, we made our move.  Now let the opponent do his, and see
       * how many eyes we can get. 
       */
      result = minimize_eyes(eye_color, &min1, &defendi1, &defendj1,
			  eyedata, eyepoints, diag_defends, diag_attacks,
			  0, NULL);
      if (result == 0) {
	if (min1 > localmin) {
	  localmin = min1;
	  localdefendi = i;
	  localdefendj = j;
 	  if (moves)
 	    location_set_clear(moves);
	}
 	if (min1 == localmin && moves)
 	  location_set_add(moves, i, j);
      }

      popgo();
    }
    assert(stackp == save_stackp);
  }
  
  if (localmin > 2)
    localmin = 2;

  *max = localmin;
  if (defendi) *defendi = localdefendi;
  if (defendj) *defendj = localdefendj;

  if (localdefendi == -1)
    eyehash_table[hashindex] |= ATTACK_NO_MOVE;
  else
    eyehash_set_defense_move(&defends, hashindex, 
			     localdefendi, localdefendj);

    eyehash_table[hashindex] |= (localmin + 1) << DEFENSE_RESULT_SHIFT;

  TRACE("exiting maximize_eyes\n");
  return 0;
}


static void
get_boundary(Location_set *boundary,
	     struct eye_data eyedata[MAX_BOARD][MAX_BOARD],
	     Location_set *eyepoints, int color)
{
  Location_set_iterator  iter;
  int  i, j;

  location_set_copy(boundary, eyepoints);
  for (ls_iter_init(boundary, &iter);
       ls_iter_next(&iter, &i, &j);
       ls_iter_update(&iter))
  {
    if (eyedata[i][j].marginal)
      location_set_del(boundary, i, j);
  }

  location_set_expand(boundary, 0);
  location_set_subtract(boundary, eyepoints);

  /* Only keep those parts of the boundary that are actually there. */
  for (ls_iter_init(boundary, &iter);
       ls_iter_next(&iter, &i, &j);
       ls_iter_update(&iter))
  {
    if (p[i][j] != color)
      location_set_del(boundary, i, j);
  }
}


/*
 * Show the board in a form suitable for debugging life code.
 * This is also callable from GDB.
 */

static void
life_showboard()
{
  int i, j;

  start_draw_board();
  for (i=0; i<board_size; i++) {
    for (j=0; j<board_size; j++) {
      int c;
      int color = p[i][j];

      if (color == WHITE) {
	if (move_in_stack(i, j, stackp_when_called))
	  c = 'o';
	else
	  c = 'O';
      }
      else if (color == BLACK) {
	if (move_in_stack(i, j, stackp_when_called))
	  c = 'x';
	else
	  c = 'X';
      }
      else
	c = EMPTY;
      draw_char(i, j, c);
    }
  }
  end_draw_board();
}

/*
 * Local Variables:
 * tab-width: 8
 * c-basic-offset: 2
 * End:
 */
