/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * This is GNU GO, a Go program. Contact gnugo@gnu.org, or see   *
 * http://www.gnu.org/software/gnugo/ for more information.      *
 *                                                               *
 * Copyright 1999 and 2000 by the Free Software Foundation.      *
 *                                                               *
 * This program is free software; you can redistribute it and/or *
 * modify it under the terms of the GNU General Public License   *
 * as published by the Free Software Foundation - version 2.     *
 *                                                               *
 * This program is distributed in the hope that it will be       *
 * useful, but WITHOUT ANY WARRANTY; without even the implied    *
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR       *
 * PURPOSE.  See the GNU General Public License in file COPYING  *
 * for more details.                                             *
 *                                                               *
 * You should have received a copy of the GNU General Public     *
 * License along with this program; if not, write to the Free    *
 * Software Foundation, Inc., 59 Temple Place - Suite 330,       *
 * Boston, MA 02111, USA                                         *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <assert.h>
#include <time.h>
#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif

#include "liberty.h"
#include "ttsgf.h"
#include "sgfana.h"
#include "interface.h"


/*
 * This function underpins all the TRACE and DEBUG stuff.
 * It accepts %c, %d and %s as usual. But it
 * also accepts %m, which takes TWO integers and writes a move
 * Nasty bodge: %o at the start means outdent, i.e. cancel indent.
 */

static void 
vgprintf(FILE* outputfile, const char *fmt, va_list ap)
{
  if (fmt[0] == '%' && fmt[1] == 'o')
    fmt += 2;  /* cancel indent */
  else if (stackp > 0)
    fprintf(outputfile, "%.*s", stackp*2, "                                ");

  for ( ; *fmt ; ++fmt ) {
    if (*fmt == '%') {
      switch (*++fmt) {
      case 'c':
      {
	/* rules of promotion => passed as int, not char */
	int c = va_arg(ap, int);
	putc(c, outputfile);
	break;
      }
      case 'd':
      {
	int d = va_arg(ap, int);
	fprintf(outputfile, "%d", d);
	break;
      }
      case 'f':
      {
	double f = va_arg(ap, double); /* passed as double, not float */
	fprintf(outputfile, "%f", f);
	break;
      }
      case 's':
      {
	char *s = va_arg(ap, char*);
	fputs(s, outputfile);
	break;
      }
      case 'm':
      {
	char movename[4];
	int m = va_arg(ap, int);
	int n = va_arg(ap, int);
	if ((m<0) || (n<0) || (m>=board_size) || (n>=board_size))
	  fputs("??", outputfile);
	else {
	  /* Generate the move name. */
	  if (n<8)
	    movename[0] = n+65;
	  else
	    movename[0] = n+66;
	  sprintf(movename+1, "%d", board_size-m);
	  fputs(movename, outputfile);
	}
	break;
      }
      default:
	fprintf(outputfile, "\n\nUnknown format character '%c'\n", *fmt);
	break;
      }
    }
    else
      putc(*fmt, outputfile);
  }
}


/*
 * required wrapper around vgprintf, writes to stderr
 */

void 
gprintf(const char *fmt, ...)
{
  va_list ap;
  va_start(ap, fmt);
  vgprintf(stderr, fmt, ap);
  va_end(ap);
}


/*
 * required wrapper around vgprintf, in contrast to gprintf this one
 * writes to stdout
 */

void
mprintf(const char *fmt, ...)
{
  va_list ap;
  va_start(ap, fmt);
  vgprintf(stdout, fmt, ap);
  va_end(ap);
}



#ifndef __GNUC__  /* gnuc allows us to define these as macros in liberty.h */

void 
TRACE(const char *fmt, ...)
{
  va_list ap;

  if (!verbose)
    return;

  va_start(ap, fmt);
  vprintf(fmt, ap);
  va_end(ap);
}


void 
VTRACE(const char *fmt, ...)
{
  va_list ap;

  if (verbose < 3)
    return;

  va_start(ap, fmt);
  vprintf(fmt, ap);
  va_end(ap);
}


void 
RTRACE(const char *fmt, ...)
{
  va_list ap;

  if (verbose < 2)
    return;

  va_start(ap, fmt);
  vprintf(fmt, ap);
  va_end(ap);
}


void 
DEBUG(int flag, const char *fmt, ...)
{
  va_list ap;

  if (!(debug & flag))
    return;

  va_start(ap, fmt);
  vprintf(fmt, ap);
  va_end(ap);
}

#endif


/*
 * A wrapper around abort() which shows the state variables at the time
 * of the problem.  (i, j) are typically a related move, or -1, -1.
 */

void 
abortgo(const char *file, int line, const char *msg, int x, int y)
{
  verbose = 4;
  TRACE("%o\n\n***assertion failure:\n%s:%d - %s near %m***\n\n",
	file, line, msg, x, y);
  dump_stack();
  showboard(0);

#if 0
  while (stackp > 0) {
    popgo();
    showboard(0);
  }
#endif

  if (sgf_root) {
    /* FIXME: How should we get seed and komi in here?
       writesgf(sgf_root, "abortgo.sgf", gameinfo->seed, gameinfo->komi); */
    sgf_write_header(sgf_root, sgf_root, analyzerflag & ANALYZE_OVERWRITE,
		     0, 0.0);
    writesgf(sgf_root, "abortgo.sgf");
  }
  
  fprintf(stderr, "\n\
gnugo %s: You stepped on a bug.\n\
Please save this game as an sgf file \
and mail together with the file abortgo.sgf to \
gnugo@gnu.org\n\n", VERSION);

  abort();  /* cause core dump */
}


/* Convert a color value to a string. */
const char *
color_to_string(int color)
{
  if (color == EMPTY)
    return "empty";
  else if (color == WHITE)
    return "white";
  else if (color == BLACK)
    return "black";
  else
    return "purple?";
}

/* ---------------------------------------------------------------- */


/*
 * See the Texinfo documentation (Dragon) for the definition of
 * strategic_distance.
 */
 
int 
strategic_distance_to(int color, int i, int j)
{
  if (color==BLACK)
    return (strategic_distance_to_black[i][j]);
  else if (color==WHITE)
    return (strategic_distance_to_white[i][j]);    
  else
    return 0;
}


/*
 * See the Texinfo documentation (Dragon) for the definition of
 * distance.
 */

int 
distance_to(int color, int i, int j)
{
  if (color==BLACK)
    return (distance_to_black[i][j]);
  else if (color==WHITE)
    return (distance_to_white[i][j]);    
  else
    return 0;
}


/*
 * Remove the string at (i,j), treating it as captured.
 * FIXME: This description doesn't look right.
 */

void
change_dragon_status(int x, int y, int status)
{
  int i,j;
  int origini = dragon[x][y].origini;
  int originj = dragon[x][y].originj;

  for (i=0; i<board_size; i++) 
    for (j=0; j<board_size; j++) {
      if (dragon[i][j].origini == origini
	  && dragon[i][j].originj == originj)
	dragon[i][j].status = status;
    }
}


/* 
 * Measure territory.
 */

void
count_territory(int *white, int *black)
{
  int i, j;

  *white = 0;
  *black = 0;
  make_worms();

  for (i=0; i<board_size; i++)
    for (j=0; j<board_size; j++)
      if (p[i][j] == EMPTY) {
	if (worm[i][j].color == BLACK_BORDER)
	  ++*black;
	else if (worm[i][j].color==WHITE_BORDER)
	  ++*white;
      }
}


/*
 * Evaluate the territory for both sides.
 * Removes dead dragons before counting.
 *
 * ATTENTION! The caller has to pushgo and popgo!
*/

void
evaluate_territory( int *white, int *black)
{
  int i, j;

  *white = 0;
  *black = 0;

  /* First we remove all dead dragons. */
  for (i=0; i<board_size; i++)
    for (j=0; j<board_size; j++) {
      if (dragon[i][j].status == DEAD
	  && p[i][j] != EMPTY)
	remove_string(i, j);
    }

  /* Make hashdata correct after removals. */
  hashdata_recalc(&hashdata, p, ko_i, ko_j);

  make_worms();
  make_dragons(EMPTY);
  for (i=0; i<board_size; i++)
    for (j=0; j<board_size; j++)
        switch(p[i][j]) {
        case EMPTY:
	  if (dragon[i][j].color == BLACK_BORDER)
	    ++*black;
	  else if (dragon[i][j].color == WHITE_BORDER)
	    ++*white;
	  break;
        case BLACK:
        case WHITE:
	  break;
      }
}


/* 
 * Change_defense(ai, aj, ti, tj, dcode) moves the point of defense of the
 * worm at (ai, aj) to (ti, tj), and sets worm[a].defend_code to dcode.
 *
 * FIXME: I think this function is obsolete in the new move generation.
 */

void
change_defense(int ai, int aj, int ti, int tj, int dcode)
{
  int origini = worm[ai][aj].origini;
  int originj = worm[ai][aj].originj;
  int si = worm[ai][aj].defendi;  /* Old defense point. */
  int sj = worm[ai][aj].defendj;
  
  assert (stackp==0);

  if (ti == -1)
    TRACE("Removed defense of %m (was %m).\n", origini, originj, si, sj);
  else if (si == -1)
    TRACE("Setting defense of %m to %m.\n", origini, originj, ti, tj);
  else
    TRACE("Moved defense of %m from %m to %m.\n", origini, originj,
	  si, sj, ti, tj);
  
  assert (ti == -1 || p[ti][tj] == EMPTY);
  assert (p[ai][aj] != EMPTY);

  if ((worm[ai][aj].attacki != -1)
      && (worm[ai][aj].defendi != -1)
      && (ti == -1)) 
  {
    int m, n;

    for (m=0; m<board_size; m++)
      for (n=0; n<board_size; n++)
	if ((dragon[m][n].origini == dragon[ai][aj].origini) 
	    && (dragon[m][n].originj == dragon[ai][aj].originj))
	  dragon[m][n].status = DEAD;
  }
  worm[origini][originj].defendi = ti;
  worm[origini][originj].defendj = tj;
  worm[origini][originj].defend_code = dcode;
  propagate_worm(origini, originj);

  if (ti != -1)
    add_defense_move(ti, tj, origini, originj);
  else if (si != -1)
    remove_defense_move(si, sj, origini, originj);
}


/*
 * change_attack(ai, aj, ti, tj, acode) moves the point of attack of the
 * worm at (ai, aj) to (ti, tj), and sets worm[a].attack_code to acode.
 */

void
change_attack(int ai, int aj, int ti, int tj, int acode)
{
  int origini = worm[ai][aj].origini;
  int originj = worm[ai][aj].originj;
  int si = worm[ai][aj].attacki;  /* Old attack point. */
  int sj = worm[ai][aj].attackj;
  
  assert(stackp==0);

  if ((ti == -1) && (si != -1)) 
    TRACE("Removed attack of %m (was %m).\n", origini, originj, si, sj);
  else if (si == -1)
    TRACE("Setting attack of %m to %m.\n", origini, originj, ti, tj);
  else
    TRACE("Moved attack of %m from %m to %m.\n", origini, originj,
	  si, sj, ti, tj);

  assert (ti == -1 || p[ti][tj] == EMPTY);
  assert (p[ai][aj] != EMPTY);

  worm[origini][originj].attacki = ti;
  worm[origini][originj].attackj = tj;
  worm[origini][originj].attack_code = acode;
  propagate_worm(origini, originj);

  if (ti != -1)
    add_attack_move(ti, tj, origini, originj);
  else if (si != -1)
    remove_attack_move(si, sj, origini, originj);
}


/*
 * Check whether a move at (ti,tj) stops the enemy from playing at (ai,aj).
 */

int
defend_against(int ti, int tj, int color, int ai, int aj)
{
  if (trymove(ti, tj, color, "defend_against", -1, -1)) {
    if (!safe_move(ai, aj, OTHER_COLOR(color))) {
      popgo();
      return 1;
    }
    popgo();
  }
  return 0;
}


/* 
 * Returns true if color can cut at (i,j). This information is
 * collected by find_cuts(), using the B patterns in the connections
 * database.
 */

int
cut_possible(int i, int j, int color)
{
  if (color == WHITE)
    return black_eye[i][j].cut;
  else
    return white_eye[i][j].cut;
}



/*
 * does_attack(ti, tj, ai, aj) returns true if the move at (ti, tj)
 * attacks (ai, aj). This means that it captures the string, and that
 * (ai, aj) is not already dead.
 */

int
does_attack(int ti, int tj, int ai, int aj)
{
  int color = p[ai][aj];
  int other = OTHER_COLOR(color);
  int result = 0;

  if ((stackp==0) 
      && (worm[ai][aj].attacki != -1) 
      && (worm[ai][aj].defendi == -1))
      return 0;
  if ((stackp>0)
      && attack(ai, aj, NULL, NULL)
      && !find_defense(ai, aj, NULL, NULL))
    return 0;
  if (trymove(ti, tj, other, "does_attack-A", ai, aj)) {
    if (p[ai][aj] && !find_defense(ai, aj, NULL, NULL)) {
      result = 1;
      increase_depth_values();
      if ((worm[ai][aj].defendi != -1) 
	  && (trymove(worm[ai][aj].defendi, worm[ai][aj].defendj, 
		      color, "does_attack-B", ai, aj))) 
      {
	if (!attack(ai, aj, NULL, NULL))
	  result = 0;
	popgo();
      }
      decrease_depth_values();
    }
    popgo();
  }

  return result;
}


/*
 * does_defend(ti, tj, ai, aj) returns true if the move at (ti, tj)
 * defends (ai, aj). This means that it defends the string, and that
 * (ai, aj) can be captured if no defense is made.
 */

int
does_defend(int ti, int tj, int ai, int aj)
{
  int color = p[ai][aj];
  int other = OTHER_COLOR(color);
  int result = 0;
  
  if ((stackp==0) && (worm[ai][aj].attacki == -1))
    return 0;
  if ((stackp>0) && !attack(ai, aj, NULL, NULL))
    return 0;
  if (trymove(ti, tj, color, "does_defend-A", ai, aj)) {
    if (!attack(ai, aj, NULL, NULL)) {
      result = 1;
      increase_depth_values();
      if ((worm[ai][aj].attacki != -1) 
	  && trymove(worm[ai][aj].attacki, worm[ai][aj].attackj, other, 
		     "does_defend-B", ai, aj)) {
	if (p[ai][aj] && !find_defense(ai, aj, NULL, NULL))
	  result = 0;
	popgo();
      }
      decrease_depth_values();
    }
    popgo();
  }

  return result;
}


/*
 * attack_either(ai, aj, bi, bj) returns true if there is a move which
 * guarantees that at least one of the strings (ai, aj) and (bi, bj)
 * can be captured. A typical application for this is in connection
 * patterns, where after a cut it suffices to capture one of the cutting
 * stones.
 *
 * FIXME: The current implementation only looks for uncoordinated
 *        attacks. This is insufficient to find double ataris or 
 *        moves such as 'a' in positions like
 *
 *        XOOOOOOOX
 *        XOXXOXXOX
 *        XX..a..XX
 *        ---------
 *
 *        where neither of the threatened X stones can be captured right
 *        out. Still either can be captured by a move down to a.
 */

int
attack_either(int ai, int aj, int bi, int bj)
{
  int color = p[ai][aj];
  ASSERT(color != EMPTY , ai, aj);
  ASSERT(color == p[bi][bj], bi, bj);

  /* Start by attacking the string with the fewest liberties. On
   * average this seems to be slightly more efficient.
   */
  if (countlib(ai, aj) <= countlib(bi, bj))
    return attack(ai, aj, NULL, NULL) || attack(bi, bj, NULL, NULL);
  else
    return attack(bi, bj, NULL, NULL) || attack(ai, aj, NULL, NULL);
}


/*
 * defend_both(ai, aj, bi, bj) returns true if both the strings (ai, aj)
 * and (bi, bj) can be defended simultaneously or if there is no attack.
 * A typical application for this is in connection patterns, where
 * after a cut it's necessary to defend both cutting stones.
 *
 * FIXME: The current implementation only looks at uncoordinated defense.
 *        A proper implementation would require some serious reading.
 */

int
defend_both(int ai, int aj, int bi, int bj)
{
  int a_threatened = 0;
  int b_threatened = 0;
  int ci, cj, di, dj;
  
  int color = p[ai][aj];
  ASSERT(color != EMPTY , ai, aj);
  ASSERT(color == p[bi][bj], bi, bj);

  if (attack(ai, aj, NULL, NULL)) {
    a_threatened = 1;
    if (!find_defense(ai, aj, &ci, &cj))
      return 0; /* (ai, aj) already lost */
  }
  
  if (attack(bi, bj, NULL, NULL)) {
    b_threatened = 1;
    if (!find_defense(bi, bj, &di, &dj))
      return 0; /* (bi, bj) already lost */
  }

  /* Neither string can be attacked or only one of them, in which case
   * we have time to save it.
   */
  if (!a_threatened || !b_threatened)
    return 1;
  
  /* If both strings are threatened we assume that one will become lost,
   * unless find_defense() happened to return the same defense point for
   * both (which e.g. may happen if they are in fact the same string).
   * This is still a bit too pessimistic, as there may be one move which
   * saves both strings. To do this right we should try each move which
   * defends either string and see if it also defends the other string.
   */

  if (ci == di && cj == dj)
    return 1; /* Both strings can be attacked but also defended by one move. */

  /* We also try each of the returned defense points and see whether
   * the other string can still be attacked. This still gives a
   * somewhat pessimistic estimation.
   */

  if (trymove(ci, cj, color, "defend_both-A", ai, aj)) {
    if (p[bi][bj] && !attack(bi, bj, NULL, NULL)) {
      popgo();
      return 1;
    }
    popgo();
  }
  
  if (trymove(di, dj, color, "defend_both-B", bi, bj)) {
    if (p[ai][aj] && !attack(ai, aj, NULL, NULL)) {
      popgo();
      return 1;
    }
    popgo();
  }
  
  /* Both strings can be attacked but we have only time to defend one. */
  return 0;
}


/*
 * break_through(ai, aj, bi, bj, ci, cj) returns 1 if a position can
 * succesfully be broken through and 2 if it can be cut. The position
 * is assumed to have the shape (the colors may be reversed)
 *
 * .O.       dbe
 * OXO       aFc
 *
 * It is X to move and try to capture at least one of a, b, and c. If
 * this succeeds, X is said to have broken through the position.
 * Otherwise X may try to cut through the position, which means
 * keeping F safe and getting a tactically safe string at either d or
 * e.
 *
 * Important notice: a, b, and c must be given in the correct order.
 *
 * FIXME: The reading involved here can most likely be improved.
 *
 * FIXME: We need to take ko results properly into account.
 */

static int
break_through_helper(int ai, int aj, int bi, int bj, int ci, int cj,
		     int di, int dj, int ei, int ej, int Fi, int Fj,
		     int color, int other);

int
break_through(int ai, int aj, int bi, int bj, int ci, int cj)
{
  int di, dj;
  int ei, ej;
  int Fi, Fj;
  int gi, gj;
  
  int color = p[ai][aj];
  int other = OTHER_COLOR(color);

  int success = 0;
  int success2 = 0;
  
  /* Basic sanity checking. */
  ASSERT(color != EMPTY , ai, aj);
  ASSERT(color == p[bi][bj], bi, bj);
  ASSERT(color == p[ci][cj], ci, cj);

  /* Construct the rest of the points in the pattern. */
  Fi = (ai + ci) / 2;    /* F midpoint between a and c. */
  Fj = (aj + cj) / 2;
  di = ai + bi - Fi;     /* Use diagonal relation a+b = d+F. */
  dj = aj + bj - Fj;
  ei = bi + ci - Fi;     /* Use diagonal relation b+c = e+F. */
  ej = bj + cj - Fj;

  /* More sanity checking. */
  ASSERT(p[di][dj] == EMPTY , di, dj);
  ASSERT(p[ei][ej] == EMPTY , ei, ej);

  /* F might already have been captured. (play_break_through_n() can't
   * detect this.
   */
  if (p[Fi][Fj] == EMPTY)
    return 0;
  
  ASSERT(p[Fi][Fj] == other , Fi, Fj);

  /* First X tries to play at d. */
  success = break_through_helper(ai, aj, bi, bj, ci, cj,
				 di, dj, ei, ej, Fi, Fj, color, other);
  if (success == 1)
    return 1;
  
  success2 = break_through_helper(ci, cj, bi, bj, ai, aj,
				  ei, ej, di, dj, Fi, Fj, color, other);

  if (success2 == 1)
    return 1;

  if (success2 == 2)
    success = 2;

  /* If we haven't been lucky yet, we might need to start by
   * defending F.
   *
   * FIXME: The function would probably be considerably faster if we
   * start by checking whether F needs defense. Beware of ko potential
   * though.
   */
  success2 = 0;
  if (attack(Fi, Fj, NULL, NULL) && find_defense(Fi, Fj, &gi, &gj)) {
    if (trymove(gi, gj, other, "break_through-A", Fi, Fj)) {
      /* Now we let O defend his position by playing either d or e.
       * FIXME: There may be other plausible moves too.
       */
      if (trymove(di, dj, color, "break_through-B", Fi, Fj)) {
	/* O connects at d, so X cuts at e. */
	if (safe_move(ei, ej, other)) {
	  success2 = 2;
	  if (!p[ci][cj] || attack(ci, cj, NULL, NULL))
	    success2 = 1;
	}
	popgo();
      }

      if (success2 > 0 && trymove(ei, ej, color, "break_through-C", Fi, Fj)) {
	/* O connects at e, so X cuts at d. */
	if (safe_move(di, dj, other)) {
	  /* success2 is already 1 or 2. */
	  if (p[ai][aj] && !attack(ai, aj, NULL, NULL))
	    success2 = 2;
	}
	else
	  success2 = 0;
	popgo();
      }
      popgo();
    }
  }
    
  if (success2 >= 1)
    return success2;

  return success;
}

/* Helper function for break_through(). Since we can symmetrically
 * start by cutting at d or e, we use the same code for both attacks,
 * simply switching positions between the two calls.
 */
static int
break_through_helper(int ai, int aj, int bi, int bj, int ci, int cj,
		     int di, int dj, int ei, int ej, int Fi, int Fj,
		     int color, int other)
{
  int success = 0;
  int gi, gj;

  if (trymove(di, dj, other, "break_through_helper-A", Fi, Fj)) {
    /* If F can be attacked we can't start in this way. */
    if (!attack(Fi, Fj, NULL, NULL)) {
      /* If d is safe too, we have at least managed to break through. */
      if (!attack(di, dj, &gi, &gj)) {
	success = 2;
	/* So now we can try to capture something. */
	if (!p[ai][aj] || !p[bi][bj] || !defend_both(ai, aj, bi, bj))
	  success = 1;
	else {
	  /* Both a and b could be defended, or didn't need to be.
	   * Let's see if a move at e is sufficient for O.
	   */
	  int attack_on_b = 0;
	  int attack_on_a = 0;

	  if (trymove(ei, ej, color, "break_through_helper-B", Fi, Fj)) {
	    if (attack(bi, bj, NULL, NULL))
	      attack_on_b = 1;
	    else if (attack(ai, aj, NULL, NULL))
	      attack_on_a = 1;
	    popgo();
	  }

	  /* Let O find a defense and play it. */
	  if (attack_on_a || attack_on_b) {
	    int hi = -1;
	    int hj = -1;
	    if (attack_on_b)
	      find_defense(ai, aj, &hi, &hj);
	    else
	      find_defense(bi, bj, &hi, &hj);
	    
	    if (hi != -1
		&& trymove(hi, hj, color, "break_through_helper-C", Fi, Fj)) {
	      /* Now we make a second cut at e, trying to capture
	       * either b or c.
	       */
	      if (trymove(ei, ej, other, "break_through_helper-D", Fi, Fj)) {
		if (!p[bi][bj] || !p[ci][cj] || !defend_both(bi, bj, ci, cj))
		  success = 1;
		popgo();
	      }
	      popgo();
	    }
	    else
	      success = 1; /* This should have been covered by
			    * defend_both(), so probably unnecessary. */
	  }
	}
      }

      /* Too bad, d could be attacked. We let O play the attack and
       * then try to make a second cut at e. But first we must test if
       * O at e is sufficient to capture d.
       */
      else {
	if (trymove(ei, ej, color, "break_through_helper-E", Fi, Fj)) {
	  if (!p[di][dj] || !find_defense(di, dj, NULL, NULL)) {
	    popgo();
	    popgo();
	    return 0;
	  }
	  popgo();
	}
	
	if (gi == ei && gj == ej) {
	  popgo();
	  return 0;
	}

	if (trymove(gi, gj, color, "break_through_helper-F", Fi, Fj)) {
	  if (trymove(ei, ej, other, "break_through_helper-G", Fi, Fj)) {
	    if (!attack(ei, ej, NULL, NULL)) {
	      success = 2;
	      if (!p[bi][bj] || !p[ci][cj] || !defend_both(bi, bj, ci, cj))
		success = 1;
	    }
	    popgo();
	  }
	  popgo();
	}
      }
	
    }
    popgo();
  }

  return success;
}


/* The function play_break_through_n() plays a sequence of moves,
 * alternating between the players and starting with color. After
 * having played through the sequence, the three last coordinate pairs
 * gives a position to be analyzed by break_through(), to see whether
 * either color has managed to enclose some stones and/or connected
 * his own stones. If any of the three last positions is empty, it's
 * assumed that the enclosure has failed, as well as the attempt to
 * connect.
 *
 * If one or more of the moves to play turns out to be illegal for
 * some reason, the rest of the sequence is played anyway, and
 * break_through() is called as if nothing special happened.
 *
 * Like break_through(), this function returns 1 if the attempt to
 * break through was succesful and 2 if it only managed to cut
 * through.
 */
   
int
play_break_through_n(int color, int num_moves, ...)
{
  va_list ap;
  int mcolor = color;
  int success = 0;
  int i;
  int played_moves = 0;
  int xi, xj;
  int yi, yj;
  int zi, zj;
  
  va_start(ap, num_moves);

  /* Do all the moves with alternating colors. */
  /* FIXME: We should really not have to push a board for each move. 
            There should be a trymove() without the call to pushgo(). */
  for (i = 0; i < num_moves; i++) {
    int ai, aj;

    ai = va_arg(ap, int);
    aj = va_arg(ap, int);

    if (ai != -1 && trymove(ai, aj, mcolor, "play_break_through_n", -1, -1))
      played_moves++;
    mcolor = OTHER_COLOR(mcolor);
  }

  /* Now do the real work. */
  xi = va_arg(ap, int);
  xj = va_arg(ap, int);
  yi = va_arg(ap, int);
  yj = va_arg(ap, int);
  zi = va_arg(ap, int);
  zj = va_arg(ap, int);
    
  /* Temporarily increase the depth values with the number of explicitly
   * placed stones.
   */
#if 0
  modify_depth_values(played_moves);
#endif
  
  if (!p[xi][xj] || !p[yi][yj] || !p[zi][zj])
    success = 1;
  else
    success = break_through(xi, xj, yi, yj, zi, zj);

#if 0
  modify_depth_values(-played_moves);
#endif
  
  /* Pop all the moves we could successfully play. */
  for (i = 0; i < played_moves; i++)
    popgo();

  va_end(ap);
  return success;
}


/* The function play_attack_defend_n() plays a sequence of moves,
 * alternating between the players and starting with color. After
 * having played through the sequence, the last coordinate pair gives
 * a target to attack or defend, depending on the value of do_attack.
 * If there is no stone present to attack or defend, it is assumed
 * that it has already been captured. If one or more of the moves to
 * play turns out to be illegal for some reason, the rest of the
 * sequence is played anyway, and attack/defense is tested as if
 * nothing special happened.
 *
 * A typical use for these functions is to set up a ladder in an
 * autohelper and see whether it works or not.
 */
   
int
play_attack_defend_n(int color, int do_attack, int num_moves, ...)
{
  va_list ap;
  int mcolor = color;
  int success = 0;
  int i;
  int played_moves = 0;
  int zi, zj;
  
  va_start(ap, num_moves);

  /* Do all the moves with alternating colors. */
  /* FIXME: We should really not have to push a board for each move. 
            There should be a trymove() without the call to pushgo(). */
  for (i = 0; i < num_moves; i++) {
    int ai, aj;

    ai = va_arg(ap, int);
    aj = va_arg(ap, int);

    if (ai != -1 && trymove(ai, aj, mcolor, "play_attack_defend_n", -1, -1))
      played_moves++;
    mcolor = OTHER_COLOR(mcolor);
  }

  /* Now do the real work. */
  zi = va_arg(ap, int);
  zj = va_arg(ap, int);

  /* Temporarily increase the depth values with the number of explicitly
   * placed stones.
   */
#if 0
  modify_depth_values(played_moves);
#endif
  
  if (do_attack) {
    if (!p[zi][zj] || attack(zi, zj, NULL, NULL))
      success = 1;
  } else {
    if (p[zi][zj]
	&& (!attack(zi, zj, NULL, NULL)
	    || find_defense(zi, zj, NULL, NULL)))
      success = 1;
  }

#if 0
  modify_depth_values(-played_moves);
#endif
  
  /* Pop all the moves we could successfully play. */
  for (i = 0; i < played_moves; i++)
    popgo();

  va_end(ap);
  return success;
}


/* The function play_attack_defend2_n() plays a sequence of moves,
 * alternating between the players and starting with color. After
 * having played through the sequence, the two last coordinate pairs
 * give two targets to simultaneously attack or defend, depending on
 * the value of do_attack. If there is no stone present to attack or
 * defend, it is assumed that it has already been captured. If one or
 * more of the moves to play turns out to be illegal for some reason,
 * the rest of the sequence is played anyway, and attack/defense is
 * tested as if nothing special happened.
 *
 * A typical use for these functions is to set up a crosscut in an
 * autohelper and see whether at least one cutting stone can be
 * captured.
 */
   
int
play_attack_defend2_n(int color, int do_attack, int num_moves, ...)
{
  va_list ap;
  int mcolor = color;
  int success = 0;
  int i;
  int played_moves = 0;
  int yi, yj;
  int zi, zj;
  
  va_start(ap, num_moves);

  /* Do all the moves with alternating colors. */
  /* FIXME: We should really not have to push a board for each move. 
            There should be a trymove() without the call to pushgo(). */
  for (i = 0; i < num_moves; i++) {
    int ai, aj;

    ai = va_arg(ap, int);
    aj = va_arg(ap, int);

    if (ai != -1 && trymove(ai, aj, mcolor, "play_attack_defend2_n", -1, -1))
      played_moves++;
    mcolor = OTHER_COLOR(mcolor);
  }

  /* Now do the real work. */
  yi = va_arg(ap, int);
  yj = va_arg(ap, int);
  zi = va_arg(ap, int);
  zj = va_arg(ap, int);

  /* Temporarily increase the depth values with the number of explicitly
   * placed stones.
   */
#if 0
  modify_depth_values(played_moves);
#endif
  
  if (do_attack) {
    if (!p[yi][yj] || !p[zi][zj] || attack_either(yi, yj, zi, zj))
      success = 1;
  } else {
    if (p[yi][yj] && p[zi][zj] && defend_both(yi, yj, zi, zj))
      success = 1;
  }

#if 0
  modify_depth_values(-played_moves);
#endif
  
  /* Pop all the moves we could successfully play. */
  for (i = 0; i < played_moves; i++)
    popgo();

  va_end(ap);
  return success;
}


/* find_lunch(m, n, *wi, *wj, *ai, *aj) looks for a worm adjoining the
 * string at (m,n) which can be easily captured. Whether or not it can
 * be defended doesn't matter.
 *
 * Returns the location of the string in (*wi, *wj), and the location
 * of the attacking move in (*ai, *aj).
 */
	
int
find_lunch(int m, int n, int *wi, int *wj, int *ai, int *aj)
{
  int i, j, vi, vj, ki, kj;

  ASSERT(p[m][n] != 0, m, n);
  ASSERT(stackp==0, m, n);

  vi = -1;
  vj = -1;
  for (i=0; i<board_size; i++)
    for (j=0; j<board_size; j++)
      if (p[i][j]==OTHER_COLOR(p[m][n]))
	if ((   i > 0
		&& worm[i-1][j].origini == worm[m][n].origini 
		&& worm[i-1][j].originj == worm[m][n].originj)
	    || (i < board_size-1
		&& worm[i+1][j].origini == worm[m][n].origini 
		&& worm[i+1][j].originj == worm[m][n].originj)
	    || (j > 0
		&& worm[i][j-1].origini == worm[m][n].origini
		&& worm[i][j-1].originj == worm[m][n].originj)
	    || (j < board_size-1
		&& worm[i][j+1].origini == worm[m][n].origini
		&& worm[i][j+1].originj == worm[m][n].originj)
	    || (i > 0 && j > 0
		&& worm[i-1][j-1].origini == worm[m][n].origini
		&& worm[i-1][j-1].originj == worm[m][n].originj)
	    || (i > 0 && j < board_size-1
		&& worm[i-1][j+1].origini == worm[m][n].origini
		&& worm[i-1][j+1].originj == worm[m][n].originj)
	    || (i < board_size-1 && j > 0
		&& worm[i+1][j-1].origini == worm[m][n].origini
		&& worm[i+1][j-1].originj == worm[m][n].originj)
	    || (i < board_size-1 && j < board_size-1
		&& worm[i+1][j+1].origini == worm[m][n].origini
		&& worm[i+1][j+1].originj == worm[m][n].originj)
	  )
	  if (((stackp==0) && (worm[i][j].attacki != -1) 
	       && (!worm[i][j].ko))
	      || ((stackp>0) && attack(i, j, &ki, &kj))) {
	    /*
	     * If several adjacent lunches are found, we pick the 
	     * juiciest. First maximize cutstone, then minimize liberties. 
	     * We can only do this if the worm data is available, 
	     * i.e. if stackp==0.
	     */
	    if ((vi) == -1 || 
		(stackp==0
		 && (worm[i][j].cutstone > worm[vi][vj].cutstone 
		     || (worm[i][j].cutstone == worm[vi][vj].cutstone 
			 && worm[i][j].liberties < worm[vi][vj].liberties)))) {
	      vi = worm[i][j].origini;
	      vj = worm[i][j].originj;
	      if (stackp>0) {
		if (ai) *ai = ki;
		if (aj) *aj = kj;
	      }
	      else {
		if (ai) *ai=worm[i][j].attacki;
		if (aj) *aj=worm[i][j].attackj;
	      }
	    }
	  }

  if (vi != -1) {
    if (wi) *wi=vi;
    if (wj) *wj=vj;
    return 1;
  }
  else
    return 0;
}


/*
 * Modify the various depth parameters. This is typically used to
 * avoid horizon effects. By temporarily increasing the depth values
 * when trying some move, one can avoid that an irrelevant move seems
 * effective just because the reading hits a depth limit earlier than
 * it did when reading only on relevant moves.
 */

void
modify_depth_values(int n)
{
  depth += n;
  backfill_depth += n;
  fourlib_depth += n;
  ko_depth += n;
}

void
increase_depth_values(void)
{
  modify_depth_values(1);
}

void
decrease_depth_values(void)
{
  modify_depth_values(-1);
}

/* These functions allow more drastic temporary modifications of the
 * depth values. Typical use is to turn certain depth values way down
 * for reading where speed is more important than accuracy, e.g. for
 * the influence function.
 */

static int save_depth;
static int save_backfill_depth;
static int save_fourlib_depth;
static int save_ko_depth;

void
set_temporary_depth_values(int d, int b, int f, int k)
{
  save_depth          = depth;
  save_backfill_depth = backfill_depth;
  save_fourlib_depth  = fourlib_depth;
  save_ko_depth       = ko_depth;

  depth          = d;
  backfill_depth = b;
  fourlib_depth  = f;
  ko_depth       = k;
}

void
restore_depth_values()
{
  depth          = save_depth;
  backfill_depth = save_backfill_depth;
  fourlib_depth  = save_fourlib_depth;
  ko_depth       = save_ko_depth;
}


/* 
 * Test whether two dragons are the same. Used by autohelpers.
 */

int
same_dragon(int ai, int aj, int bi, int bj)
{
  return (dragon[ai][aj].origini == dragon[bi][bj].origini
	  && dragon[ai][aj].originj == dragon[bi][bj].originj);
}


/*
 * Test whether two worms are the same. Used by autohelpers.
 */

int
same_worm(int ai, int aj, int bi, int bj)
{
  return (worm[ai][aj].origini == worm[bi][bj].origini
	  && worm[ai][aj].originj == worm[bi][bj].originj);
}


/*
 * Test whether the origin of the worm at (wi, wj) is (i, j).
 */

int
is_worm_origin(int wi, int wj, int i, int j)
{
  return (worm[wi][wj].origini == i
	  && worm[wi][wj].originj == j);
}


/* Play a stone at (m, n) and count the number of liberties for the
 * resulting string. This requires (m, n) to be empty.
 *
 * This function differs from approxlib() by the fact that it removes
 * captured stones before counting the liberties.
 */

int
accurate_approxlib(int m, int n, int color, int maxlib, int *libi, int *libj)
{
  int libs = 0;

  assert(p[m][n] == EMPTY);
  assert(color != EMPTY);

  /* Use tryko() since we don't care whether the move would violate
   * the ko rule.
   */
  if (tryko(m, n, color, NULL)) {
    if (libi != NULL)
      libs = findlib(m, n, maxlib, libi, libj);
    else
      libs = countlib(m, n);
    popgo();
  }
  
  return libs;
}


/* This function will detect some blunders. If the move reduces the
 * number of liberties of an adjacent friendly string, there is a
 * danger that the move could backfire, so the function checks that no
 * friendly worm which was formerly not attackable becomes attackable,
 * and it checks that no opposing worm which was not defendable
 * becomes defendable. Only worms with worm.size>size are checked.
 *
 * For use when called from fill_liberty, this function may optionally
 * return a point of defense, which, if taken, will presumably make
 * the move at (i, j) safe on a subsequent turn.
 */

int
confirm_safety(int i, int j, int color, int size, int *di, int *dj)
{
  int libi[5], libj[5];
  int libs = accurate_approxlib(i, j, color, 5, libi, libj);
  int other = OTHER_COLOR(color);
  int issafe = 1;
  int m, n;
  
  if (di) *di = -1;
  if (dj) *dj = -1;

  TRACE("Checking safety of a %s move at %m\n", color_to_string(color), i, j);
  if (libs>4)
    return 1;

  /* Need to increase the depth values during this reading to avoid
   * horizon effects.
   */
  increase_depth_values();
  
  if ((    (i>0)
	   && (p[i-1][j] == color)
	   && (libs < worm[i-1][j].liberties))
       || ((i<board_size-1)
	   && (p[i+1][j] == color)
	   && (libs < worm[i+1][j].liberties))
       || ((j>0)
	   && (p[i][j-1] == color)
	   && (libs < worm[i][j-1].liberties))
       || ((j<board_size-1)
	   && (p[i][j+1] == color)
	   && (libs < worm[i][j+1].liberties))) {
    if (trymove(i, j, color, NULL, -1, -1)) {
      for (m=0; m<board_size; m++)
	for (n=0; n<board_size; n++)
	  if (issafe
	      && p[m][n]
	      && (worm[m][n].origini==m)
	      && (worm[m][n].originj==n)) {
	    if ((p[m][n]==color)
		&& (worm[m][n].attacki == -1) 
		&& (worm[m][n].size > size)
		&& attack(m, n, NULL, NULL)) {
	      issafe = 0;
	      TRACE("Worm at %m becomes attackable.\n", m, n);
	    }
	    else if ((p[m][n]==other)
		     && (worm[m][n].attacki != -1)
		     && (worm[m][n].defendi == -1)
		     && (worm[m][n].size > size)
		     && find_defense(m, n, NULL, NULL)) {
	      issafe = 0;
	      TRACE("Worm at %m becomes defendable.\n", m, n);
	    }
	  }
      if (libs==2) {
	if (double_atari(libi[0], libj[0], other)) {
	  if (di && safe_move(libi[0], libj[0], color)) {
	    *di = libi[0];
	    *dj = libj[0];
	  }
	  issafe = 0;
	  TRACE("Double threat appears at %m.\n", libi[0], libj[0]);
	}
	else if (double_atari(libi[1], libj[1], other)) {
	  if (di && safe_move(libi[1], libj[1], color)) {
	    *di = libi[1];
	    *dj = libj[1];
	  }
	  issafe = 0;
	  TRACE("Double threat appears at %m.\n", libi[1], libj[1]);
	}
      }
      popgo();
    }
  }
  
  /* Reset the depth values. */
  decrease_depth_values();
  
  return issafe;
}


/* Returns true if a move by (color) fits the following shape:
 * 
 *     .
 *    X*.       (O=color)
 *    OX
 * 
 * capturing one of the two X strings. The name is a slight
 * misnomer since this includes attacks which are not necessarily
 * double ataris, though the common double atari is the most
 * important special case.
 */

int
double_atari(int m, int n, int color)
{
  int other=OTHER_COLOR(color);
  int k;

  if ((m==0) || (m==board_size-1) || (n==0) || (n==board_size-1))
    return 0;

  /* Loop over the diagonal directions. */
  for (k=4; k<8; k++) {
    int dm = deltai[k];
    int dn = deltaj[k];
    
    if ((p[m+dm][n+dn]==color)
	&& (p[m][n+dn]==other)
	&& (p[m+dm][n]==other)
	&& (p[m-dm][n]==EMPTY)
	&& (p[m][n-dn]==EMPTY)
	&& trymove(m, n, color, "double_atari", -1, -1)) {
      if (!defend_both(m, n+dn, m+dm, n)) {
	popgo();
	return 1;
      }
      popgo();
    }
  }
  
  return 0;
}
    


/* This function returns true if it is judged that the capture of the
 * string at (m,n) is sufficient to create one eye. The current
 * just checks that (m,n) is not a singleton on the first line.
 */

int
vital_chain(int m, int n)
{
  if ((singleton(m, n)
       && ((m == 0) 
	   || (m == board_size-1) 
	   || (n == 0) 
	   || (n == board_size-1))))
      return 0;
  else return 1;
}


/* Get the time of day, calling gettimeofday from sys/time.h
 * if available, otherwise substituting a workaround for portability.
 */

double
gg_gettimeofday(void)
{
#ifdef HAVE_GETTIMEOFDAY
  struct timeval tv;
  gettimeofday(&tv, NULL);
  return tv.tv_sec + 1.e-6*tv.tv_usec;
#else
  return time(NULL);
#endif
}


/* Evaluates the eyespace at (m,n) and prints a report */

void
decideeye(int m, int n)
{
  int  color;
  int  max, min;
  int  attacki, attackj;
  int  defendi, defendj;
  int  i, j;
  int  save_verbose = verbose;

  initialize_engine();
  verbose = 0;
  examine_position(BLACK, EXAMINE_DRAGONS);
  verbose = save_verbose;
  
  if (black_eye[m][n].color == BLACK_BORDER) 
    color = BLACK;
  else if (white_eye[m][n].color == WHITE_BORDER) 
    color = WHITE;
  else {
    gprintf("The eye at %m is not of a single color.\n", m, n);
    return;
  }

  if (printboard)
    showboard(0);

  if (black_eye[m][n].color == BLACK_BORDER) {
    i = black_eye[m][n].origini;
    j = black_eye[m][n].originj;
    compute_eyes(i, j, &max, &min, &attacki, &attackj, &defendi, &defendj,
		 black_eye, 0, EMPTY);
    if (terse) {
      gprintf("%d,%d", min, max);
      if (max != min) {
	gprintf(":%m,%m", attacki, attackj, defendi, defendj);
      } 
      gprintf("\n");
    } 
    else {
      gprintf("Black eyespace at %m: min=%d, max=%d\n", i, j, min, max);
      if (max != min) {
	gprintf("  vital points: %m (attack) %m (defense)\n", attacki, attackj,
		defendi, defendj);
      }
    }
  }
  if (white_eye[m][n].color == WHITE_BORDER) {
    i = white_eye[m][n].origini;
    j = white_eye[m][n].originj;
    compute_eyes(i, j, &max, &min, &attacki, &attackj, &defendi, &defendj,
		 white_eye, 0, EMPTY);
    if (terse) {
      gprintf("%d,%d", min, max);
      if (max != min) {
	gprintf(":%m,%m", attacki, attackj, defendi, defendj);
      } 
      gprintf("\n");
    } 
    else {
      gprintf("White eyespace at %m: min=%d, max=%d\n", i, j, min, max);
      if (max != min) {
	gprintf("  vital points: %m (attack) %m (defense)\n", attacki, attackj,
		defendi, defendj);
      }
    }
  }
}


/* Computes the number of eyes yielded by capturing a lunch.
 * The surrounding liberties are filled and the stones are
 * removed from the board. Then compute_eyes is called to
 * evaluate the resulting eyespace. The maximum and minimum
 * number of resulting eyes is returned in the variables
 * *max and *min.
 * 
 * FIXME: The function does not try to decide whether
 * capturing the lunch results in escape.
 */

int
sniff_lunch(int i, int j, int *max, int *min)
{
  int color, other, libi[MAXLIBS], libj[MAXLIBS], libs, k;
  char mw[MAX_BOARD][MAX_BOARD];
  static struct eye_data lunch_eye[MAX_BOARD][MAX_BOARD];
  static struct half_eye_data lunch_half_eye[MAX_BOARD][MAX_BOARD];
  int m, n;

  color = p[i][j];
  other = OTHER_COLOR(color);
  libs = findlib(i, j, MAXLIBS, libi, libj);
  
  if (trymove(-1, -1, color, "sniff_lunch", i, j)) {
    for (k=0; k<libs; k++)
      p[libi[k]][libj[k]] = other;
    memset(mw, 0, sizeof(mw));
    remove_string_and_mark(i, j, mw, 1);
    originate_eye(i, j, i, j, NULL, NULL, lunch_eye);
    for (m=0; m<board_size; m++)
      for (n=0; n<board_size; n++)
	lunch_half_eye[m][n].type = 0;
    for (m=0; m<board_size; m++)
      for (n=0; n<board_size; n++)
	if (mw[m][n]) {
	  int ai, aj, di, dj;
	  int sum = topological_eye(m, n, other, &ai, &aj, &di, &dj);
	  if (sum <= 4) {
	    lunch_half_eye[m][n].type = FALSE_EYE;
	    if ((lunch_eye[m][n].esize == 1) || legal(m, n, other))
	      add_half_eye(m, n, lunch_eye, half_eye);
	    else if (sum == 3) {
	      lunch_half_eye[m][n].type = HALF_EYE;
#if 0
	  /* See comment in add_halfeye() */
	      lunch_half_eye[m][n].ai = ai;
	      lunch_half_eye[m][n].aj = aj;
	      lunch_half_eye[m][n].di = di;
	      lunch_half_eye[m][n].dj = dj;
	      add_half_eye(m, n, lunch_eye, lunch_half_eye);
#endif
	    }
	  }
	}	    
    compute_eyes(i, j, max, min, NULL, NULL, NULL, NULL, lunch_eye, 0, EMPTY);
    
    popgo();
    return 1;
  }
  else return 0;
}


/* Find those worms of the given color that can never be captured,
 * even if the opponent is allowed an arbitrary number of consecutive
 * moves. The coordinates of the origins of these worms are written to
 * the wormi, wormj arrays and the number of non-capturable worms is
 * returned.
 *
 * The algorithm is to cycle through the worms until none remains or
 * no more can be captured. A worm is removed when it is found to be
 * capturable, by letting the opponent try to play on all its
 * liberties. If the attack fails, the moves are undone.
 *
 * Notice: This function is not yet in use and has not been tested.
 */

int
unconditional_life(int wormi[MAX_STRINGS], int wormj[MAX_STRINGS], int color)
{
  int something_captured = 1; /* To get into the first turn of the loop. */
  int moves_played = 0;
  int save_moves;
  int m, n;
  int k;
  int libi[MAXLIBS];
  int libj[MAXLIBS];
  int libs;
  int other = OTHER_COLOR(color);
  int number_of_strings;
  
  while (something_captured) {
    /* Nothing captured so far in this turn of the loop. */
    something_captured = 0;

    /* Visit all friendly strings on the board. */
    for (m=0; m<board_size; m++)
      for (n=0; n<board_size; n++) {
	if (p[m][n] != color || !is_worm_origin(m, n, m, n))
	  continue;
	
	/* Try to capture the worm at (m, n). */
	libs = findlib(m, n, MAXLIBS, libi, libj);
	save_moves = moves_played;
	for (k=0; k<libs; k++) {
	  if (trymove(libi[k], libj[k], other, "unconditional_life", m, n))
	    moves_played++;
	}

	/* Successful if already captured or a single liberty remains.
	 * Otherwise we must rewind and take back the last batch of moves.
	 */
	if (p[m][n] == EMPTY)
	  something_captured = 1;
	else if (findlib(m, n, 2, libi, libj) == 1) {
	  /* Need to use tryko as a defense against the extreme case
           * when the only opponent liberty that is not suicide is an
           * illegal ko capture, like in this 5x5 position:
	   * +-----+
	   * |.XO.O|
	   * |XXOO.|
	   * |X.XOO|
	   * |XXOO.|
	   * |.XO.O|
	   * +-----+
	   */
	  int success = tryko(libi[0], libj[0], other, "unconditional_life");
	  assert(success);
	  moves_played++;
	  something_captured++;
	}
	else
	  while (moves_played > save_moves) {
	    popgo();
	    moves_played--;
	  }
      }
  }

  /* The strings still remaining are uncapturable. */
  number_of_strings = 0;
  for (m=0; m<board_size; m++)
    for (n=0; n<board_size; n++)
      if (p[m][n] == color && is_worm_origin(m, n, m, n)) {
	wormi[number_of_strings] = m;
	wormj[number_of_strings] = n;
	number_of_strings++;
      }

  /* Take back all moves. */
  while (moves_played > 0) {
    popgo();
    moves_played--;
  }

  return number_of_strings;
}


/*
 * Local Variables:
 * tab-width: 8
 * c-basic-offset: 2
 * End:
 */
