/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * This is GNU GO, a Go program. Contact gnugo@gnu.org, or see   *
 * http://www.gnu.org/software/gnugo/ for more information.      *
 *                                                               *
 * Copyright 1999 and 2000 by the Free Software Foundation.      *
 *                                                               *
 * This program is free software; you can redistribute it and/or *
 * modify it under the terms of the GNU General Public License   *
 * as published by the Free Software Foundation - version 2.     *
 *                                                               *
 * This program is distributed in the hope that it will be       *
 * useful, but WITHOUT ANY WARRANTY; without even the implied    *
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR       *
 * PURPOSE.  See the GNU General Public License in file COPYING  *
 * for more details.                                             *
 *                                                               *
 * You should have received a copy of the GNU General Public     *
 * License along with this program; if not, write to the Free    *
 * Software Foundation, Inc., 59 Temple Place - Suite 330,       *
 * Boston, MA 02111, USA                                         *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>

#include "gnugo.h"
#include "interface.h"
#include "liberty.h"
#include "gtp.h"

#define DECLARE(func) static int func(Gameinfo *gameinfo, char *s, int id)

DECLARE(gtp_playblack);
DECLARE(gtp_playwhite);
DECLARE(gtp_what_color);
DECLARE(gtp_countlib);
DECLARE(gtp_findlib);
DECLARE(gtp_quit);
DECLARE(gtp_loadsgf);
DECLARE(gtp_showboard);
DECLARE(gtp_attack);
DECLARE(gtp_defend);
DECLARE(gtp_mouse_attack);
DECLARE(gtp_mouse_defend);


/* List of known commands. */
static struct gtp_command commands[] = {
  {"black",         gtp_playblack},
  {"white",         gtp_playwhite},
  {"color",         gtp_what_color},
  {"countlib",      gtp_countlib},
  {"findlib",       gtp_findlib},
  {"quit",          gtp_quit},
  {"loadsgf",       gtp_loadsgf},
  {"showboard",     gtp_showboard},
  {"attack",        gtp_attack},
  {"defend",        gtp_defend},
  {"mouse_attack",  gtp_mouse_attack},
  {"mouse_defend",  gtp_mouse_defend},
  {NULL,            NULL}
};


/* Start playing using the Go Text Protocol. */
void
play_gtp(Gameinfo *gameinfo)
{
  /* Try to make sure that we have a useful level of buffering of stdout. */
#ifdef HAVE_SETLINEBUF
  setlinebuf(stdout);
#else
  setbuf(stdout, NULL);
#endif

  /* Prepare pattern matcher and reading code. */
  initialize_engine();

  gtp_main_loop(gameinfo, commands);
}


/****************************
 * Administrative commands. *
 ****************************/

/* Function:  Quit
 * Arguments: none
 * Fails:     never
 * Returns:   nothing
 */
static int
gtp_quit(Gameinfo *gameinfo, char *s, int id)
{
  UNUSED(gameinfo);
  UNUSED(s);
  gtp_success(id, "");
  return GTP_QUIT;
}


/******************
 * Playing moves. *
 ******************/

/* Function:  Play a black stone at the given vertex.
 * Arguments: vertex
 * Fails:     invalid vertex
 * Returns:   nothing
 */
static int
gtp_playblack(Gameinfo *gameinfo, char *s, int id)
{
  int i, j;
  if (!gtp_decode_coord(s, gameinfo->position.boardsize, &i, &j))
    return gtp_failure(id, "invalid coordinate");

  position_play_move(&gameinfo->position, i, j, BLACK);
  return gtp_success(id, "");
}


/* Function:  Play a white stone at the given vertex.
 * Arguments: vertex
 * Fails:     invalid vertex
 * Returns:   nothing
 */
static int
gtp_playwhite(Gameinfo *gameinfo, char *s, int id)
{
  int i, j;
  if (!gtp_decode_coord(s, gameinfo->position.boardsize, &i, &j))
    return gtp_failure(id, "invalid coordinate");

  position_play_move(&gameinfo->position, i, j, WHITE);
  return gtp_success(id, "");
}


/* Function:  Load an sgf file, possibly up to a move number or the first
 *            occurence of a move.           
 * Arguments: filename + move number, vertex, or nothing
 * Fails:     missing filename or failure to open or parse file
 * Returns:   nothing
 */
static int
gtp_loadsgf(Gameinfo *gameinfo, char *s, int id)
{
  char filename[GTP_BUFSIZE];
  SGFNode *sgf;
  int n;
  if (sscanf(s, "%s %n", filename, &n) != 1)
    return gtp_failure(id, "missing filename");
  
  if ((sgf = readsgffile(filename)) == NULL)
    return gtp_failure(id, "cannot open or parse '%s'", filename);

  ginfo_clear(gameinfo, 19);
  load_sgf_header(sgf, gameinfo);
  load_sgf_file(sgf, gameinfo, s+n);
  return gtp_success(id, "");
}


/*****************
 * Board status. *
 *****************/

/* Function:  Return the color at a vertex.
 * Arguments: vertex
 * Fails:     invalid vertex
 * Returns:   "black", "white", or "empty"
 */
static int
gtp_what_color(Gameinfo *gameinfo, char *s, int id)
{
  int i, j;
  if (!gtp_decode_coord(s, gameinfo->position.boardsize, &i, &j))
    return gtp_failure(id, "invalid coordinate");
  
  return gtp_success(id, color_to_string(p[i][j]));
}


/* Function:  Count number of liberties for the string at a vertex.
 * Arguments: vertex
 * Fails:     invalid vertex, empty vertex
 * Returns:   Number of liberties.
 */
static int
gtp_countlib(Gameinfo *gameinfo, char *s, int id)
{
  int i, j;
  if (!gtp_decode_coord(s, gameinfo->position.boardsize, &i, &j))
    return gtp_failure(id, "invalid coordinate");

  if (p[i][j] == EMPTY)
    return gtp_failure(id, "vertex must not be empty");

  return gtp_success(id, "%d", countlib(i, j));
}


/* Function:  Return the positions of the liberties for the string at a vertex.
 * Arguments: vertex
 * Fails:     invalid vertex, empty vertex
 * Returns:   Sorted space separated list of vertices.
 */
static int
gtp_findlib(Gameinfo *gameinfo, char *s, int id)
{
  int i, j;
  int libi[MAXLIBS];
  int libj[MAXLIBS];
  int libs;
  
  if (!gtp_decode_coord(s, gameinfo->position.boardsize, &i, &j))
    return gtp_failure(id, "invalid coordinate");

  if (p[i][j] == EMPTY)
    return gtp_failure(id, "vertex must not be empty");

  libs = findlib(i, j, MAXLIBS, libi, libj);
  gtp_printid(id, GTP_SUCCESS);
  gtp_print_vertices(libs, libi, libj, board_size);
  gtp_printf("\n\n");
  return GTP_OK;
}


/*********************
 * Tactical reading. *
 *********************/

/* Function:  Try to attack a string.
 * Arguments: vertex
 * Fails:     invalid vertex, empty vertex
 * Returns:   attack code followed by attack point if attack code nonzero.
 */
static int
gtp_attack(Gameinfo *gameinfo, char *s, int id)
{
  int i, j;
  int ai, aj;
  int attack_code;
  
  if (!gtp_decode_coord(s, gameinfo->position.boardsize, &i, &j))
    return gtp_failure(id, "invalid coordinate");

  if (p[i][j] == EMPTY)
    return gtp_failure(id, "vertex must not be empty");

  attack_code = attack(i, j, &ai, &aj);
  gtp_printid(id, GTP_SUCCESS);
  gtp_printf("%d", attack_code);
  if (attack_code > 0) {
    gtp_printf(" ");
    gtp_print_vertex(ai, aj, board_size);
  }
  gtp_printf("\n\n");
  return GTP_OK;
}  


/* Function:  Try to defend a string.
 * Arguments: vertex
 * Fails:     invalid vertex, empty vertex
 * Returns:   defense code followed by defense point if defense code nonzero.
 */
static int
gtp_defend(Gameinfo *gameinfo, char *s, int id)
{
  int i, j;
  int di, dj;
  int defend_code;
  
  if (!gtp_decode_coord(s, gameinfo->position.boardsize, &i, &j))
    return gtp_failure(id, "invalid coordinate");

  if (p[i][j] == EMPTY)
    return gtp_failure(id, "vertex must not be empty");

  defend_code = find_defense(i, j, &di, &dj);
  gtp_printid(id, GTP_SUCCESS);
  gtp_printf("%d", defend_code);
  if (defend_code > 0) {
    gtp_printf(" ");
    gtp_print_vertex(di, dj, board_size);
  }
  gtp_printf("\n\n");
  return GTP_OK;
}  


/******************
 * mouse reading. *
 ******************/

static char goal[MAX_BOARD][MAX_BOARD];

/* Function:  Try to attack a dragon.
 * Arguments: vertex
 * Fails:     invalid vertex, empty vertex
 * Returns:   attack code followed by attack point if attack code nonzero.
 */
static int
gtp_mouse_attack(Gameinfo *gameinfo, char *s, int id)
{
  int i, j;
  int ai, aj;
  int attack_code;
  int save_verbose = verbose;
  if (!gtp_decode_coord(s, gameinfo->position.boardsize, &i, &j))
    return gtp_failure(id, "invalid coordinate");

  if (p[i][j] == EMPTY)
    return gtp_failure(id, "vertex must not be empty");

  verbose=0;
  examine_position(p[i][j], EXAMINE_DRAGONS);
  verbose=save_verbose;
  
  attack_code = mouse_attack(i, j, &ai, &aj, goal);
  gtp_printid(id, GTP_SUCCESS);
  gtp_printf("%d", attack_code);
  if (attack_code > 0) {
    gtp_printf(" ");
    gtp_print_vertex(ai, aj, board_size);
  }
  gtp_printf("\n\n");
  return GTP_OK;
}  


/* Function:  Try to defend a dragon.
 * Arguments: vertex
 * Fails:     invalid vertex, empty vertex
 * Returns:   defense code followed by defense point if defense code nonzero.
 */
static int
gtp_mouse_defend(Gameinfo *gameinfo, char *s, int id)
{
  int i, j;
  int di, dj;
  int defend_code;
  int save_verbose = verbose;
  
  if (!gtp_decode_coord(s, gameinfo->position.boardsize, &i, &j))
    return gtp_failure(id, "invalid coordinate");

  if (p[i][j] == EMPTY)
    return gtp_failure(id, "vertex must not be empty");

  verbose=0;
  examine_position(p[i][j], EXAMINE_DRAGONS);
  verbose=save_verbose;

  defend_code = mouse_defend(i, j, &di, &dj, goal);
  gtp_printid(id, GTP_SUCCESS);
  gtp_printf("%d", defend_code);
  if (defend_code > 0) {
    gtp_printf(" ");
    gtp_print_vertex(di, dj, board_size);
  }
  gtp_printf("\n\n");
  return GTP_OK;
}  


/*********
 * debug *
 *********/

/* Function:  Write the position to stderr.
 * Arguments: none
 * Fails:     never
 * Returns:   nothing
 */
static int
gtp_showboard(Gameinfo *gameinfo, char *s, int id)
{
  UNUSED(gameinfo);
  UNUSED(s);
  showboard(0);
  return gtp_success(id, "");
}

/*
 * Local Variables:
 * tab-width: 8
 * c-basic-offset: 2
 * End:
 */
