/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * This is GNU GO, a Go program. Contact gnugo@gnu.org, or see   *
 * http://www.gnu.org/software/gnugo/ for more information.      *
 *                                                               *
 * Copyright 1999 and 2000 by the Free Software Foundation.      *
 *                                                               *
 * This program is free software; you can redistribute it and/or *
 * modify it under the terms of the GNU General Public License   *
 * as published by the Free Software Foundation - version 2.     *
 *                                                               *
 * This program is distributed in the hope that it will be       *
 * useful, but WITHOUT ANY WARRANTY; without even the implied    *
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR       *
 * PURPOSE.  See the GNU General Public License in file COPYING  *
 * for more details.                                             *
 *                                                               *
 * You should have received a copy of the GNU General Public     *
 * License along with this program; if not, write to the Free    *
 * Software Foundation, Inc., 59 Temple Place - Suite 330,       *
 * Boston, MA 02111, USA                                         *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "interface.h"

#include "sgf_properties.h"
#include "sgf_utils.h"
#include "sgfana.h"
#include "gnugo.h"
#include "random.h"


void
play_solo(int boardsize, int moves)
{
  Gameinfo  gameinfo;

  int passes = 0; /* num. consecutive passes */
  int to_move = BLACK;
  int move_val;
  int mycolor;
  int yourcolor;
  double t1, t2;
  int benchmark = moves;
  
  /* It tends not to be very imaginative in the opening,
   * so we scatter a few stones randomly to start with.
   * We add two random numbers to reduce the probability
   * of playing stones near the edge.
   */
  
  int n = 6 + 2*gg_rand()%5;
  int i, j;

  int handicap;
  int seed;
  float komi;
  
  ginfo_clear(&gameinfo, boardsize);

  komi     = gameinfo.komi;
  seed     = gameinfo.seed;
  handicap = gameinfo.handicap;
  mycolor  = gameinfo.computer_player;
  
  yourcolor = OTHER_COLOR(mycolor);
  
  sgffile_write_gameinfo(&gameinfo, "solo");
  
  /* Generate some random moves. */
  if (boardsize > 6) {
    do {
      do {
	i = (gg_rand() % 4) + (gg_rand() % (boardsize - 4));
	j = (gg_rand() % 4) + (gg_rand() % (boardsize - 4));
      } while (gameinfo.position.board[i][j] != EMPTY);
      
      position_play_move(&gameinfo.position, i, j, to_move);
      sgffile_move_made(i, j, to_move, 0);
      sgfAddPlay(0, to_move, i, j);
      sgfAnalyzer(0);
      to_move = OTHER_COLOR(to_move);
    } while (--n > 0);
  }
  
  t1 = gg_gettimeofday();
  
  while (passes < 2 && --moves >= 0 && !time_to_die) {
    move_val = position_genmove(&gameinfo.position, &i, &j, to_move,
				gameinfo.move_number);
    position_play_move(&gameinfo.position, i, j, to_move);
    if (move_val < 0) {
      ++passes;
      printf("%s(%d): Pass\n", to_move==BLACK ? "Black" : "White",
	     gameinfo.move_number);
    } else {
      passes = 0;
      gameinfo.move_number++;
      gprintf("%s(%d): %m\n", to_move==BLACK ? "Black" : "White",
	      gameinfo.move_number, i, j);
    }
    sgffile_move_made(i, j, to_move, move_val);
    sgfAddPlay(0, to_move, i, j);
    sgfAnalyzer(move_val);
    to_move=OTHER_COLOR(to_move);
  }
  t2 = gg_gettimeofday();
  
  /* Two passes and it's over. */
  who_wins(mycolor, komi, stdout);
  if (t2 == t1) {
    printf("%.3f moves played\n", (double) (benchmark-moves));
  } else {
    printf("%.3f moves/sec\n", (benchmark-moves)/(t2-t1));
  }
}


/* ================================================================ */


/* FIXME : this should be in a separate source file, but it
 * is easier to send diffs when there are no new source
 * files to add.
 */


/* This used to read the sgf file character by character,
 * but now we have a full sgf parser available. Variations
 * are not supported.
 *
 * The name is historical: The file has already been read into a
 * tree. Walk the main variation, actioning the properties into the
 * playing board.
 *
 * Returns the color of the next move to be made.
 *
 * Head is a sgf tree (already read from file). until is an optional
 * string of the form either 'L12' or '120' which tells it to stop
 * loading at that move or move-number - when debugging, this will be
 * a particularly bad move, and we want to know why.
 */

int
load_sgf_file(SGFNode *head, Gameinfo *gameinfo, const char *untilstr)
{
  int bs, handicap;
  int next = BLACK;
  
  int untilm = -1, untiln = -1;
  int until = 9999;
  int addstone = 0;          /* handicap stone detector */
  
  SGFNode *node;
  
  if (sgfGetIntProperty(head, "SZ", &bs))
    position_clear(&gameinfo->position, bs);
  else
    position_clear(&gameinfo->position, 19);

  /* Now we can safely parse the until string (which depends on board size). */
  
  if (untilstr) {
    if (*untilstr > '0' && *untilstr <= '9') {
      until = atoi(untilstr);
      DEBUG(DEBUG_LOADSGF, "Loading until move %d\n", until);
    }
    else {
      untiln = *untilstr - 'A';
      if (*untilstr >= 'I')
	--untiln;
	  
      untilm = gameinfo->position.boardsize - atoi(untilstr+1);
      DEBUG(DEBUG_LOADSGF, "Loading until move at %d,%d (%m)\n", 
	    untilm, untiln, untilm, untiln);
    }
  }
  
  if (sgfGetIntProperty(head, "HA", &handicap) && handicap > 0) {
    gameinfo->handicap = handicap;
    next = WHITE;
  }
  
  /* Finally, we iterate over all the properties of all the
   * nodes, actioning them. We follow only the 'child' pointers,
   * as we have no interest in variations.
   *
   * The sgf routines map AB[aa][bb][cc] into AB[aa]AB[bb]AB[cc]
   */
  for (node=head; node; node=node->child) {
    SGFProperty *prop;
    int i, j;
      
    sgfSetLastNode(node);

    for (prop=node->props; prop; prop=prop->next) {
      DEBUG(DEBUG_LOADSGF, "%c%c[%s]\n", 
	    prop->name & 0xff, (prop->name >> 8), prop->value);
      switch(prop->name) {
      case SGFAB:
	get_moveXY(prop, &i, &j, gameinfo->position.boardsize);
	put_stone(&gameinfo->position, i, j, BLACK);
	addstone = 1;
	break;
	      
      case SGFAW:
	get_moveXY(prop, &i, &j, gameinfo->position.boardsize);
	put_stone(&gameinfo->position, i, j, WHITE);
	addstone = 1;
	break;
	      
      case SGFPL:
	/* following really should not be needed for proper sgf file */
	if (!gameinfo->move_number && !addstone)
	  position_sethand(&gameinfo->position, gameinfo->handicap, 0);
	      
	if (prop->value[0] == 'w' || prop->value[0] == 'W')
	  next = WHITE;
	else
	  next = BLACK;
	break;
	      
      case SGFW:
      case SGFB:
	next = prop->name == SGFW ? WHITE : BLACK;
	/* following really should not be needed for proper sgf file */
	if (!gameinfo->move_number && !addstone)
	  position_sethand(&gameinfo->position, gameinfo->handicap, 0);
	gameinfo->move_number++;
	if (gameinfo->move_number == until)
	  return next;
	      
	if (get_moveXY(prop, &i, &j, gameinfo->position.boardsize)) {
	  if (i == untilm && j == untiln)
	    return next;
		  
	  put_move(&gameinfo->position, i, j, next);
	  next = OTHER_COLOR(next);
	  /* send the move to the file, move_val=0 */
	  /*  sgf_move_made(i, j, next, 0);        */
	  /* play_move(i, j, next);                */
	}
	      
	break;
      }
    }
  }

  return next;
}


/*
 * Reads header info from sgf structure and sets the appropriate variables.
 */

void
load_sgf_header(SGFNode *head, Gameinfo *gameinfo)
{
  int bsize;
  int handicap;
  float komi;
  
  if (sgfGetIntProperty(head, "SZ", &bsize))
    position_clear(&gameinfo->position, bsize);
  else
    position_clear(&gameinfo->position, 19);
  
  if (sgfGetIntProperty(head, "HA", &handicap) && handicap > 0)
    /* Handicap stones should appear as AW,AB properties in the sgf file. */
    gameinfo->handicap = handicap;

  if (sgfGetFloatProperty(head, "KM", &komi))
    gameinfo->komi = komi;
}


/*
 * Load SGF file and run genmove().
 */

void 
load_and_analyze_sgf_file(SGFNode *head, Gameinfo *gameinfo, 
			  const char *untilstr, int benchmark)
{
  int i, j;
  int next;
  
  /* In a position from a file, probably both players have played. */
  color_has_played = WHITE | BLACK;

  /* We load the header to get correct board_size, komi, and handicap
   * for writing.
   */
  

  load_sgf_header(head, gameinfo); 
  sgffile_write_gameinfo(gameinfo, "load and analyze");
  next = load_sgf_file(head, gameinfo, untilstr);
  
  if (benchmark) {
    int z;
    for (z=0; z < benchmark; ++z) {
      genmove(&i, &j, next);
      next = OTHER_COLOR(next);
    }
  }
  else {
    genmove(&i, &j, next);
    
    if (!is_pass(i, j)) {
      gprintf("%s move %m\n", next == WHITE ? "white (o)" : "black (X)", i,j);
      put_move(&gameinfo->position, i, j, next);
    }
    else {
      gprintf("%s move : PASS!\n", next == WHITE ? "white (o)" : "black (X)");
      sgffile_move_made(i, j, next, 0);
    }
  }
}


/*
 * Load SGF file and score the game
 * untilstr:
 * end  - finish the game by selfplaying from the end of the file until 
 *        two passes
 * last - estimate territorial balance at the end of the of the file
 * move - load file until move is reached and estimate territorial balance
 */

void 
load_and_score_sgf_file(SGFNode *head, Gameinfo *gameinfo, 
			const char *untilstr)
{
  int i, j, move_val, pass, wt, bt;
  int until;
  float score;
  float komi;
  float result;
  char *tempc = NULL;
  char dummy;
  char text[250];
  char winner;
  int next;
  pass = 0;

  assert(head);
  assert(sgf_root);
  load_sgf_header(head, gameinfo);
  sgffile_write_gameinfo(gameinfo, "load and score");
  next = load_sgf_file(head, gameinfo, untilstr);
  komi = gameinfo->komi;
  
  until = atoi(untilstr);
  if (!strcmp(untilstr, "end")) {
    until = 9999;
    do {
      move_val = genmove(&i, &j, next);
      play_move(i, j, next);
      if (move_val >= 0) {
	pass = 0;
	gprintf("%d %s move %m\n", gameinfo->move_number,
		next == WHITE ? "white (o)" : "black (X)", i, j);
      } else {
	++pass;
	gprintf("%d %s move : PASS!\n", gameinfo->move_number, 
		next == WHITE ? "white (o)" : "black (X)");
      }
      sgffile_move_made(i, j, next, move_val);
      sgfAddPlay(0, next, i, j);
      sgfAnalyzer(move_val);
      gameinfo->move_number++;
      next = OTHER_COLOR(next);
    } while ((gameinfo->move_number <= until) && (pass < 2));

    if (pass >= 2) {
      /*count actually owned territory*/
      evaluate_territory(&wt, &bt);
      score = gameinfo->position.white_captured + bt 
	- gameinfo->position.black_captured - wt - komi;
      if (score>0) {
	sprintf(text, "Black wins by %1.1f points\n", score);
	winner = 'B';
      }
      else if (score<0) {
	sprintf(text, "White wins by %1.1f points\n", -1.0 * score);
	winner = 'W';
      }
      else {
	sprintf(text, "Tie\n");
	winner = '0';
      }
      fputs(text, stdout);
      sgfShowTerritory(0);
      sgfAddComment(NULL, text);
      sgffile_write_comment(text);
      if (sgfGetCharProperty(head, "RE", &tempc)) {
	if (sscanf(tempc, "%1c%f", &dummy, &result)==2) {
	  fprintf(stdout, "Result from file: %1.1f\n", result);
	  fputs("GNU Go result and result from file are ", stdout);
	  if ((result==abs(score)) && (winner==dummy))
	    fputs("identical\n", stdout);
	  else
	    fputs("different\n", stdout);
	      
	} else {
	  if (tempc[2] == 'R') {
	    fprintf(stdout, "Result from file: Resign\n");
	    fputs("GNU Go result and result from file are ", stdout);
	    if (tempc[0] == winner)
	      fputs("identical\n", stdout);
	    else
	      fputs("different\n", stdout);
	  }
	}
      }
      sgfWriteResult(score);
      return;
    }
  }

  move_val = genmove(&i, &j, next);
  sgfAnalyzer(move_val);
  score = terri_eval[BLACK] - terri_eval[WHITE] - komi;
  sprintf(text,
"Black moyo: %3i\n\
White moyo: %3i\n\
Black territory: %3i\n\
White territory: %3i\n\
Dead white stones: %3i\n\
Dead black stones: %3i\n\
%s seems to win by %1.1f points\n",
	  moyo_eval[BLACK], moyo_eval[WHITE],
	  terri_eval[BLACK], terri_eval[WHITE],
	  gameinfo->position.white_captured,
	  gameinfo->position.black_captured,
	  (score>0) ? ("BLACK") : ((score<0) ? ("WHITE") : ("NOBODY")),
	  (score<0) ? (-1*score) : (score));
  if (strcmp(untilstr, "last"))
      sgfSetLastNode(sgfStartVariant(0));
  sgfShowMoyo(SGF_SHOW_COLOR, 0);
  sgfAddComment(NULL, text);
  sgffile_write_comment(text);
  fputs(text, stdout);
}


/*
 * Local Variables:
 * tab-width: 8
 * c-basic-offset: 2
 * End:
 */
